from energizer.neural_network import Module, Parameter
from energizer.tensor import Tensor
import mlx.core as mx
import numpy as np


class Embedding(Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: int = None,
        max_norm: float = None,
        norm_type: float = 2.0,
        scale_grad_by_freq: bool = False,
        sparse: bool = False,
        device: str = "cpu",
    ):
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        self.max_norm = max_norm
        self.norm_type = norm_type
        self.scale_grad_by_freq = scale_grad_by_freq
        self.sparse = sparse
        self.device = device

        if device == "gpu":
            self.weight = Parameter(
                mx.random.normal(shape=(num_embeddings, embedding_dim)),
                requires_grad=True,
                device=device,
            )
        else:
            self.weight = Parameter(
                np.random.randn(num_embeddings, embedding_dim),
                requires_grad=True,
                device=device,
            )

        if padding_idx is not None:
            if padding_idx < 0 or padding_idx >= num_embeddings:
                raise ValueError(
                    f"padding_idx must be in the range [0, {num_embeddings})"
                )
            self._zero_padding_idx()

    def _zero_padding_idx(self):
        if self.device == "gpu":
            weight_data = self.weight.data
            weight_data[self.padding_idx] = mx.zeros(self.embedding_dim)
            self.weight.data = weight_data
        else:
            self.weight.data[self.padding_idx] = np.zeros(self.embedding_dim)

    def _apply_max_norm(self, weight: Tensor):
        if self.max_norm is not None:
            return

        if self.device == "gpu":
            weight_norm = mx.linalg.norm(self.weight.data, ord=self.norm_type, axis=1)
            scale = mx.minimum(1.0, self.max_norm / (weight_norm + 1e-8))
            self.weight.data = self.weight.data * scale[:, mx.newaxis]
        else:
            weight_norm = np.linalg.norm(
                self.weight.data, ord=self.norm_type, axis=1, keepdims=True
            )
            scale = np.minimum(1.0, self.max_norm / (weight_norm + 1e-8))
            self.weight.data = self.weight.data * scale

    def forward(self, x: Tensor) -> Tensor:
        if self.max_norm is not None and self.training:
            # apply norm
            self._apply_max_norm(self.weight.data)

        if hasattr(self, "weight"):
            indices = x.data
        else:
            indices = x

        if self.device == "gpu":
            if not isinstance(indices, mx.array):
                indices = mx.array(indices)
            embedded = mx.take(self.weight.data, indices, axis=0)

            if self.padding_idx is not None and self.training:
                mask = mx.not_equal(indices, self.padding_idx)
                if self.scale_grad_by_freq:
                    pass
        else:
            embedded = self.weight.data[indices]

            if self.padding_idx is not None and self.training:
                mask = indices != self.padding_idx
                self._last_input_mask = mask

        output = Tensor(embedded, requires_grad=x.requires_grad, device=x.device)

        if self.training:
            self._last_input_mask = mask

        return output

    def extra_repr(self):
        return f"num_embeddings={self.num_embeddings}, embedding_dim={self.embedding_dim}, padding_idx={self.padding_idx}, max_norm={self.max_norm}, norm_type={self.norm_type}, scale_grad_by_freq={self.scale_grad_by_freq}, sparse={self.sparse}"



from __future__ import annotations
from typing import Any
import energizer.tensor as ts
import numpy as np

try:
    import mlx.core as mx
except ImportError:
    mx = None


def _reduce_grad_to_shape(grad: Any, shape: tuple) -> Any:
    if shape == ():
        if mx and isinstance(grad, mx.array):
            return mx.sum(grad)
        return np.sum(grad)

    is_mlx = mx and isinstance(grad, mx.array)

    g = grad
    while len(getattr(g, "shape", ())) > len(shape):
        if is_mlx:
            g = mx.sum(g, axis=0)
        else:
            g = np.sum(g, axis=0)

    for axis, (gdim, tdim) in enumerate(zip(g.shape, shape)):
        if tdim == 1 and gdim != 1:
            if is_mlx:
                g = mx.sum(g, axis=axis, keepdims=True)
            else:
                g = np.sum(g, axis=axis, keepdims=True)

    try:
        return g.reshape(shape)
    except Exception:
        return g


def mul_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    def get_data(t):
        """Extract raw data from tensor or return as-is"""
        if isinstance(t, (int, float)):
            return t
        elif isinstance(t, ts.Tensor):
            return t.data
        else:
            return t

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        other_data = get_data(tensors[1])
        if mx and (isinstance(grad_data, mx.array) or isinstance(other_data, mx.array)):
            grad_mx = (
                grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            )
            other_mx = (
                other_data
                if isinstance(other_data, mx.array)
                else (
                    mx.array(other_data)
                    if isinstance(other_data, np.ndarray)
                    else other_data
                )
            )
            tensors[0].grad = grad_mx * other_mx
        else:
            tensors[0].grad = grad_data * other_data
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        other_data = get_data(tensors[0])
        if mx and (isinstance(grad_data, mx.array) or isinstance(other_data, mx.array)):
            grad_mx = (
                grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            )
            other_mx = (
                other_data
                if isinstance(other_data, mx.array)
                else (
                    mx.array(other_data)
                    if isinstance(other_data, np.ndarray)
                    else other_data
                )
            )
            tensors[1].grad = grad_mx * other_mx
        else:
            tensors[1].grad = grad_data * other_data
        tensors[1].backward(tensors[1].grad)
    else:
        return grad_data * get_data(tensors[1])
    return grad_data * get_data(tensors[0])


def add_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad0 = (
            grad_outputs[0].data
            if isinstance(grad_outputs[0], ts.Tensor)
            else grad_outputs[0]
        )
        tensors[0].grad = _reduce_grad_to_shape(grad0, tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        grad1 = (
            grad_outputs[0].data
            if isinstance(grad_outputs[0], ts.Tensor)
            else grad_outputs[0]
        )
        tensors[1].grad = _reduce_grad_to_shape(grad1, tensors[1].data.shape)
        tensors[1].backward(tensors[1].grad)
    else:
        return grad_outputs[0]
    return grad_outputs[0]


def sub_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad0 = (
            grad_outputs[0].data
            if isinstance(grad_outputs[0], ts.Tensor)
            else grad_outputs[0]
        )
        tensors[0].grad = _reduce_grad_to_shape(grad0, tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        grad1 = (
            grad_outputs[0].data
            if isinstance(grad_outputs[0], ts.Tensor)
            else grad_outputs[0]
        )
        tensors[1].grad = _reduce_grad_to_shape(-grad1, tensors[1].data.shape)
        tensors[1].backward(tensors[1].grad)
    else:
        return -grad_outputs[0]
    return grad_outputs[0]


def neg_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = (
            -grad_outputs[0]
            if isinstance(grad_outputs[0], (int, float))
            else -grad_outputs[0].data
        )
        tensors[0].backward(tensors[0].grad)
    else:
        return (
            -grad_outputs[0]
            if isinstance(grad_outputs[0], (int, float))
            else -grad_outputs[0].data
        )
    return grad_outputs[0]


def truediv_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    is_reverse_div = isinstance(tensors[0], (int, float)) and isinstance(
        tensors[1], ts.Tensor
    )

    if is_reverse_div:
        if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
            tensors[1].grad = -grad_data * tensors[0] / (tensors[1].data ** 2)
            tensors[1].backward(tensors[1].grad)
        return grad_data

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data / (
            tensors[1] if isinstance(tensors[1], (int, float)) else tensors[1].data
        )
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        tensors[1].grad = -grad_data * tensors[0].data / (tensors[1].data ** 2)
        tensors[1].backward(tensors[1].grad)
    return grad_data


def matmul_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    def ensure_compatible(a, b):
        if mx and isinstance(a, mx.array) and not isinstance(b, mx.array):
            return a, mx.array(b)
        elif mx and isinstance(b, mx.array) and not isinstance(a, mx.array):
            return mx.array(a), b
        return a, b

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        grad_compat, weight_T = ensure_compatible(grad_data, tensors[1].data.T)
        tensors[0].grad = grad_compat @ weight_T
        tensors[0].backward(tensors[0].grad)
    if isinstance(tensors[1], ts.Tensor) and tensors[1].requires_grad:
        input_T, grad_compat = ensure_compatible(tensors[0].data.T, grad_data)
        tensors[1].grad = input_T @ grad_compat
        tensors[1].backward(tensors[1].grad)

    grad_compat, weight_T = ensure_compatible(grad_data, tensors[1].data.T)
    return grad_compat @ weight_T


def pow_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * tensors[1] * (tensors[0].data ** (tensors[1] - 1))
        tensors[0].backward(tensors[0].grad)
    return grad_data * tensors[1] * (tensors[0].data ** (tensors[1] - 1))


def max_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * (tensors[0].data > tensors[1])
        tensors[0].backward(tensors[0].grad)
    return grad_data * (tensors[0].data > tensors[1])


def sum_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        shape = tensors[0].data.shape
        if mx and isinstance(tensors[0].data, mx.array):
            g = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            tensors[0].grad = mx.broadcast_to(g, shape)
        else:
            if not isinstance(grad_data, np.ndarray):
                grad_data = np.array(grad_data)
            tensors[0].grad = np.broadcast_to(grad_data, shape).copy()
        tensors[0].backward(tensors[0].grad)
    return grad_data


def mean_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        n = tensors[0].data.size
        shape = tensors[0].data.shape
        if mx and isinstance(tensors[0].data, mx.array):
            g = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
            tensors[0].grad = mx.broadcast_to(g / n, shape)
        else:
            if not isinstance(grad_data, np.ndarray):
                grad_data = np.array(grad_data)
            tensors[0].grad = np.broadcast_to(grad_data / n, shape).copy()
        tensors[0].backward(tensors[0].grad)
    return grad_data / tensors[0].data.size


def sigmoid_backward(tensors: Any, grad_outputs: Any) -> Any:
    """
    Backward for sigmoid(x).

    `tensors` is expected to be [x, y] where y = sigmoid(x) (raw array).
    """
    x = tensors[0]
    y = tensors[1]
    grad = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    # dy/dx = y * (1 - y)
    if mx and (isinstance(grad, mx.array) or isinstance(y, mx.array)):
        grad_mx = grad if isinstance(grad, mx.array) else mx.array(grad)
        y_mx = y if isinstance(y, mx.array) else mx.array(y)
        grad_x = grad_mx * (y_mx * (1 - y_mx))
    else:
        grad_np = np.array(grad) if not isinstance(grad, np.ndarray) else grad
        y_np = np.array(y) if not isinstance(y, np.ndarray) else y
        grad_x = grad_np * (y_np * (1 - y_np))

    if isinstance(x, ts.Tensor) and x.requires_grad:
        x.grad = grad_x
        x.backward(x.grad)

    return grad_x


def as_strided_backward(tensors: Any, grad_outputs: Any) -> Any:
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = np.lib.stride_tricks.as_strided(
            grad_outputs[0],
            shape=tensors[0].data.shape,
            strides=tensors[1].data.strides,
            storage_offset=tensors[2].data,
        )
        tensors[0].backward(tensors[0].grad)
    return grad_outputs[0]


def trace_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        n = min(tensors[0].data.shape[-2:])
        if mx and isinstance(tensors[0].data, mx.array):
            tensors[0].grad = mx.eye(n) * grad_data
        else:
            tensors[0].grad = np.eye(n) * np.asarray(grad_data)
        tensors[0].backward(tensors[0].grad)
    return grad_data


def conv1d_backward(
    tensors: Any, grad_outputs: Any, stride: int = 1, padding: int = 0
) -> Any:
    x, weight, bias = tensors
    grad_output = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    if not isinstance(grad_output, np.ndarray):
        grad_output = np.array(grad_output)

    weight_data = weight.data if isinstance(weight, ts.Tensor) else weight

    if isinstance(x, ts.Tensor) and x.requires_grad:
        grad_x = conv1d_grad_input(grad_output, weight_data, stride, padding)
        if grad_x is not None:
            x.grad = grad_x
            x.backward(x.grad)

    if isinstance(weight, ts.Tensor) and weight.requires_grad:
        grad_weight = conv1d_grad_weight(grad_output, x, stride, padding)
        if grad_weight is not None:
            weight.grad = grad_weight
            weight.backward(weight.grad)

    if bias is not None and isinstance(bias, ts.Tensor) and bias.requires_grad:
        grad_bias = grad_output.sum(axis=(0, 2))
        bias.grad = grad_bias
        bias.backward(bias.grad)

    return grad_output


def conv1d_grad_input(
    grad_output: np.ndarray, weight: "ts.Tensor", stride: int = 1, padding: int = 0
) -> np.ndarray:
    batch_size, out_channels, out_length = grad_output.shape
    _, in_channels, kernel_size = weight.data.shape

    length = (out_length - 1) * stride + kernel_size - 2 * padding

    grad_input_padded = np.zeros((batch_size, in_channels, length + 2 * padding))

    for b in range(batch_size):
        for oc in range(out_channels):
            for ol in range(out_length):
                il_start = ol * stride

                for k in range(kernel_size):
                    grad_input_padded[b, :, il_start + k] += (
                        weight[oc, :, k] * grad_output[b, oc, ol]
                    )

    if padding > 0:
        return grad_input_padded[:, :, padding:-padding]
    return grad_input_padded


def conv1d_grad_weight(
    grad_output: np.ndarray, x: "ts.Tensor", stride: int = 1, padding: int = 0
) -> np.ndarray:
    batch_size, out_channels, out_length = grad_output.shape
    _, in_channels, _ = x.data.shape

    kernel_size = x.data.shape[2] - (out_length - 1) * stride + 2 * padding

    grad_weight = np.zeros((out_channels, in_channels, kernel_size))

    if padding > 0:
        x_padded = np.pad(x.data, ((0, 0), (0, 0), (padding, padding)), mode="constant")
    else:
        x_padded = x.data

    for b in range(batch_size):
        for oc in range(out_channels):
            for ol in range(out_length):
                il_start = ol * stride
                il_end = il_start + kernel_size

                x_window = x_padded[b, :, il_start:il_end]

                grad_weight[oc] += x_window * grad_output[b, oc, ol]

    return grad_weight


def reshape_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if mx and isinstance(grad_data, mx.array):
            tensors[0].grad = mx.reshape(grad_data, tensors[0].data.shape)
        else:
            if not isinstance(grad_data, np.ndarray):
                grad_data = np.array(grad_data)
            tensors[0].grad = grad_data.reshape(tensors[0].data.shape)
        tensors[0].backward(tensors[0].grad)

    if mx and isinstance(grad_data, mx.array):
        return mx.reshape(grad_data, tensors[0].data.shape)
    else:
        if not isinstance(grad_data, np.ndarray):
            grad_data = np.array(grad_data)
        return grad_data.reshape(tensors[0].data.shape)


def getitem_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if mx and isinstance(tensors[0].data, mx.array):
            g = np.zeros(tensors[0].data.shape, dtype=np.float32)
            g[tensors[1]] = np.array(grad_data)
            tensors[0].grad = mx.array(g)
        else:
            tensors[0].grad = np.zeros_like(tensors[0].data)
            tensors[0].grad[tensors[1]] = grad_data
        tensors[0].backward(tensors[0].grad)
    if mx and isinstance(tensors[0].data, mx.array):
        return mx.zeros_like(tensors[0].data)
    return np.zeros_like(tensors[0].data)


def setitem_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if tensors[0].grad is None:
            if mx and isinstance(tensors[0].data, mx.array):
                tensors[0].grad = mx.zeros_like(tensors[0].data)
            else:
                tensors[0].grad = np.zeros_like(tensors[0].data)
        if mx and isinstance(tensors[0].grad, mx.array):
            g = np.array(tensors[0].grad)
            g[tensors[1]] = np.array(grad_data)
            tensors[0].grad = mx.array(g)
        else:
            tensors[0].grad[tensors[1]] = grad_data
        tensors[0].backward(tensors[0].grad)
    if mx and isinstance(tensors[0].data, mx.array):
        return mx.zeros_like(tensors[0].data)
    return np.zeros_like(tensors[0].data)


def item_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data
        tensors[0].backward(tensors[0].grad)
    return grad_data


def transpose_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        if len(tensors) == 3:
            dim0, dim1 = tensors[1], tensors[2]
            if mx and isinstance(grad_data, mx.array):
                tensors[0].grad = mx.swapaxes(grad_data, dim0, dim1)
            else:
                tensors[0].grad = np.swapaxes(grad_data, dim0, dim1)
        else:
            if mx and isinstance(grad_data, mx.array):
                tensors[0].grad = mx.transpose(grad_data)
            else:
                tensors[0].grad = grad_data.T
        tensors[0].backward(tensors[0].grad)

    if len(tensors) == 3:
        dim0, dim1 = tensors[1], tensors[2]
        if mx and isinstance(grad_data, mx.array):
            return mx.swapaxes(grad_data, dim0, dim1)
        return np.swapaxes(grad_data, dim0, dim1)
    if mx and isinstance(grad_data, mx.array):
        return mx.transpose(grad_data)
    return grad_data.T


def tanh_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * (1 - tensors[0].data ** 2)
        tensors[0].backward(tensors[0].grad)
    return grad_data * (1 - tensors[0].data ** 2)


def softmax_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    x_data = tensors[0].data if isinstance(tensors[0], ts.Tensor) else tensors[0]

    if mx and isinstance(x_data, mx.array):
        s = mx.softmax(x_data, axis=-1)
        g = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
        grad_x = s * (g - mx.sum(g * s, axis=-1, keepdims=True))
    else:
        if not isinstance(x_data, np.ndarray):
            x_data = np.array(x_data)
        e_x = np.exp(x_data - x_data.max(axis=-1, keepdims=True))
        s = e_x / e_x.sum(axis=-1, keepdims=True)
        g = grad_data if isinstance(grad_data, np.ndarray) else np.array(grad_data)
        grad_x = s * (g - (g * s).sum(axis=-1, keepdims=True))

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_x
        tensors[0].backward(tensors[0].grad)
    return grad_x


def gelu_backward(tensors: Any, grad_outputs: Any) -> Any:
    """
    GELU(x) = 0.5 * x * (1 + tanh(sqrt(2/π) * (x + 0.044715 * x³)))

    d(GELU)/dx = 0.5 * (1 + tanh(t)) + 0.5 * x * (1 − tanh(t)²) * t'
    where  t  = sqrt(2/π) * (x + 0.044715 * x³)
    and    t' = sqrt(2/π) * (1 + 3 * 0.044715 * x²)
    """
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    x_data = tensors[0].data if isinstance(tensors[0], ts.Tensor) else tensors[0]

    if mx and isinstance(x_data, mx.array):
        g = grad_data if isinstance(grad_data, mx.array) else mx.array(grad_data)
        c = mx.array(np.sqrt(2.0 / np.pi))
        t = c * (x_data + 0.044715 * mx.power(x_data, 3))
        tanh_t = mx.tanh(t)
        t_prime = c * (1.0 + 3.0 * 0.044715 * x_data**2)
        grad_x = g * (0.5 * (1.0 + tanh_t) + 0.5 * x_data * (1.0 - tanh_t**2) * t_prime)
    else:
        x_np = x_data if isinstance(x_data, np.ndarray) else np.array(x_data)
        g = grad_data if isinstance(grad_data, np.ndarray) else np.array(grad_data)
        c = np.sqrt(2.0 / np.pi)
        t = c * (x_np + 0.044715 * np.power(x_np, 3))
        tanh_t = np.tanh(t)
        t_prime = c * (1.0 + 3.0 * 0.044715 * x_np**2)
        grad_x = g * (0.5 * (1.0 + tanh_t) + 0.5 * x_np * (1.0 - tanh_t**2) * t_prime)

    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_x
        tensors[0].backward(tensors[0].grad)
    return grad_x


def dropout_backward(tensors: Any, grad_outputs: Any) -> Any:
    """
    tensors = [x, mask]
    mask is the scaled binary mask produced during the forward pass
    (already multiplied by 1 / (1 - p)).
    Backward is simply grad * mask — same operation as the forward, no new randomness.
    """
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    x = tensors[0]
    mask = tensors[1]  # raw numpy array

    if mx and isinstance(grad_data, mx.array):
        m = mx.array(mask) if not isinstance(mask, mx.array) else mask
        grad_x = grad_data * m
    else:
        g = grad_data if isinstance(grad_data, np.ndarray) else np.array(grad_data)
        grad_x = g * mask

    if isinstance(x, ts.Tensor) and x.requires_grad:
        x.grad = grad_x
        x.backward(x.grad)
    return grad_x


def cross_entropy_backward(tensors: Any, grad_outputs: Any) -> Any:
    """
    Cross-entropy from raw logits.
    tensors = [logits_tensor, target_tensor]
      logits : (B, C)  raw (pre-softmax) scores
      target : (B,)    integer class indices   OR   (B, C) one-hot / soft targets

    Gradient of mean cross-entropy w.r.t. logits:
        (softmax(logits) − one_hot(target)) / B
    """
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    logits = tensors[0]
    target = tensors[1]

    logits_data = logits.data if isinstance(logits, ts.Tensor) else logits
    target_data = target.data if isinstance(target, ts.Tensor) else target

    logits_np = np.array(logits_data).astype(np.float32)
    target_np = np.array(target_data)

    B = logits_np.shape[0]
    shifted = logits_np - logits_np.max(axis=1, keepdims=True)
    softmax_probs = np.exp(shifted)
    softmax_probs /= softmax_probs.sum(axis=1, keepdims=True)

    grad_logits = softmax_probs.copy()
    if target_np.ndim == 1:
        grad_logits[np.arange(B), target_np.astype(int)] -= 1.0
    else:
        grad_logits -= target_np.astype(np.float32)
    grad_logits /= B

    # Scale by the upstream scalar gradient (typically 1.0 for a scalar loss)
    upstream = float(np.array(grad_data).flat[0])
    grad_logits = (grad_logits * upstream).astype(np.float32)

    if mx and isinstance(logits_data, mx.array):
        grad_arr = mx.array(grad_logits)
    else:
        grad_arr = grad_logits

    if isinstance(logits, ts.Tensor) and logits.requires_grad:
        logits.grad = grad_arr
        logits.backward(logits.grad)
    return grad_arr


def squeeze_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data
        tensors[0].backward(tensors[0].grad)
    return grad_data


def exp_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * tensors[0].data
        tensors[0].backward(tensors[0].grad)
    return grad_data * tensors[0].data


def log_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data / tensors[0].data
        tensors[0].backward(tensors[0].grad)
    return grad_data / tensors[0].data


def clamp_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = (
            grad_data
            * (tensors[0].data >= tensors[1])
            * (tensors[0].data <= tensors[2])
        )
        tensors[0].backward(tensors[0].grad)
    return grad_data * (tensors[0].data >= tensors[1]) * (tensors[0].data <= tensors[2])


def minimum_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * (tensors[0].data <= tensors[1])
        tensors[0].backward(tensors[0].grad)
    return grad_data * (tensors[0].data <= tensors[1])


def maximum_backward(tensors: Any, grad_outputs: Any) -> Any:
    grad_data = (
        grad_outputs[0].data
        if isinstance(grad_outputs[0], ts.Tensor)
        else grad_outputs[0]
    )
    if isinstance(tensors[0], ts.Tensor) and tensors[0].requires_grad:
        tensors[0].grad = grad_data * (tensors[0].data >= tensors[1])
        tensors[0].backward(tensors[0].grad)
    return grad_data * (tensors[0].data >= tensors[1])



#!/usr/bin/env python3
"""
BACKPACK — Apple Silicon Power Monitor
Tracks CPU · GPU · ANE power in real time via powermetrics.

Usage:
    sudo python3 backpack_monitor.py
    sudo python3 backpack_monitor.py -i 250

Keybinds:
    q   Quit
    p   Pause / Resume
    r   Reset peaks
    h   Cycle history window
    +   Faster sampling
    -   Slower sampling
"""

import os, re, sys, tty, signal, select, argparse
import subprocess, threading, time, termios
from collections import deque
from typing import Optional

# ── ANSI helpers ──────────────────────────────────────────────────────────────

R = "\033[0m"


def fg(r, g, b):
    return f"\033[38;2;{r};{g};{b}m"


def bg(r, g, b):
    return f"\033[48;2;{r};{g};{b}m"


C_CPU = fg(100, 180, 255)
C_GPU = fg(200, 120, 255)
C_ANE = fg(80, 220, 140)
C_WARN = fg(255, 200, 60)
C_HOT = fg(255, 80, 80)
C_DIM = fg(80, 90, 105)
C_WHITE = fg(220, 225, 235)

ANSI_RE = re.compile(r"\033\[[^m]*m|\033\][^\a]*\a")


def vlen(s: str) -> int:
    return len(ANSI_RE.sub("", s))


def rpad(s: str, width: int, fill: str = " ") -> str:
    return s + fill * max(0, width - vlen(s))


def center_str(s: str, width: int) -> str:
    v = vlen(s)
    pad = max(0, width - v)
    return " " * (pad // 2) + s + " " * (pad - pad // 2)


SPARKS = " ▁▂▃▄▅▆▇█"


def sparkline(hist: deque, width: int, col: str) -> str:
    pts = list(hist)[-width:] if hist else []
    mx = max(pts) if pts else 1
    if mx == 0:
        mx = 1
    out = C_DIM + "·" * max(0, width - len(pts)) + R
    for v in pts:
        idx = max(0, min(len(SPARKS) - 1, int((v / mx) * (len(SPARKS) - 1))))
        out += col + SPARKS[idx] + R
    return out


def power_bar(value: int, peak: int, width: int, col: str) -> str:
    if peak <= 0:
        peak = 1
    pct = max(0.0, min(1.0, value / peak))
    filled = int(width * pct)
    color = col if pct < 0.5 else (C_WARN if pct < 0.8 else C_HOT)
    return color + "█" * filled + R + C_DIM + "░" * (width - filled) + R


def fmt_energy(mj: float) -> str:
    if mj >= 1_000_000:
        return f"{mj/1_000_000:.2f}kJ"
    if mj >= 1_000:
        return f"{mj/1_000:.2f}J "
    return f"{mj:.0f}mJ"


def fmt_elapsed(s: float) -> str:
    m, s = divmod(int(s), 60)
    h, m = divmod(m, 60)
    return f"{h:02d}:{m:02d}:{s:02d}" if h else f"{m:02d}:{s:02d}"


# ── State ─────────────────────────────────────────────────────────────────────

HIST_WINDOWS = [60, 120, 300]


class State:
    def __init__(self, interval_ms: int):
        self.power = {"CPU": 0, "GPU": 0, "ANE": 0}
        self.peak = {"CPU": 0, "GPU": 0, "ANE": 0}
        self.history = {k: deque(maxlen=6000) for k in ("CPU", "GPU", "ANE")}
        self.energy = {"CPU": 0.0, "GPU": 0.0, "ANE": 0.0}
        self.paused = False
        self.interval = interval_ms
        self.hist_idx = 0
        self.start = time.time()
        self.lock = threading.Lock()

    @property
    def hist_window(self):
        return HIST_WINDOWS[self.hist_idx % len(HIST_WINDOWS)]

    def record(self, key: str, val: int):
        if self.paused:
            return
        with self.lock:
            self.power[key] = val
            self.peak[key] = max(self.peak[key], val)
            self.history[key].append(val)
            self.energy[key] += val * (self.interval / 1000.0)

    def rolling_avg(self, key: str) -> float:
        with self.lock:
            h = list(self.history[key])
        pts = int(self.hist_window * 1000 / max(self.interval, 1))
        recent = h[-pts:] if h else []
        return sum(recent) / len(recent) if recent else 0.0

    def reset_peaks(self):
        with self.lock:
            for k in self.peak:
                self.peak[k] = self.power[k]


# ── Render ────────────────────────────────────────────────────────────────────

LOGO = [
    "  ╔╗ ╔═╗╔═╗╦╔═╔═╗╔═╗╔═╗╦╔═",
    "  ╠╩╗╠═╣║  ╠╩╗╠═╝╠═╣║  ╠╩╗",
    "  ╚═╝╩ ╩╚═╝╩ ╩╩  ╩ ╩╚═╝╩ ╩",
]
LOGO_COLS = [fg(60, 160, 255), fg(80, 200, 255), fg(130, 220, 200)]

COMP = {
    "CPU": (C_CPU, "◈"),
    "GPU": (C_GPU, "◉"),
    "ANE": (C_ANE, "◆"),
}


def render(state: State) -> str:
    try:
        W = max(60, os.get_terminal_size().columns)
    except OSError:
        W = 100

    out = []
    add = out.append

    def rule(ch="─"):
        add(C_DIM + ch * W + R)

    # ── Logo ──────────────────────────────────────────────────────────────────
    add("")
    for i, row in enumerate(LOGO):
        add(center_str(LOGO_COLS[i] + "\033[1m" + row + R, W))
    add("")

    # ── Status line ───────────────────────────────────────────────────────────
    elapsed = fmt_elapsed(time.time() - state.start)
    live = C_WARN + "\033[1m ⏸ PAUSED " + R if state.paused else C_ANE + "● LIVE" + R
    status = (
        C_DIM
        + f"  ⏱ {elapsed}   "
        + R
        + C_DIM
        + f"window \033[1m{state.hist_window}s\033[0m   "
        + R
        + C_DIM
        + f"interval \033[1m{state.interval}ms\033[0m   "
        + R
        + live
    )
    add(rpad(status, W))
    rule("─")

    # ── Component sections ────────────────────────────────────────────────────
    BAR_W = max(10, W - 4)  # 2 spaces each side

    with state.lock:
        snap_pw = dict(state.power)
        snap_pk = dict(state.peak)
        snap_en = dict(state.energy)
        snap_hi = {k: deque(v) for k, v in state.history.items()}

    for key in ("CPU", "GPU", "ANE"):
        col, icon = COMP[key]
        val = snap_pw[key]
        peak = snap_pk[key]
        avg = state.rolling_avg(key)
        pct = val / peak if peak > 0 else 0.0
        pcol = col if pct < 0.5 else (C_WARN if pct < 0.8 else C_HOT)

        # Metric row
        label = col + "\033[1m" + f"  {icon} {key}" + R
        v_str = pcol + "\033[1m" + f"  {val:>5} mW" + R
        p_str = pcol + f"  {pct*100:>5.1f}%" + R
        k_str = C_DIM + f"   pk {peak:>5} mW" + R
        a_str = C_DIM + f"  avg {avg:>5.0f} mW" + R
        e_str = C_DIM + f"  Σ {fmt_energy(snap_en[key])}" + R

        add(rpad(label + v_str + p_str + k_str + a_str + e_str, W))

        # Bar
        add("  " + power_bar(val, peak, BAR_W, col))

        # Sparkline
        pts_in_window = int(state.hist_window * 1000 / max(state.interval, 1))
        add("  " + sparkline(snap_hi[key], min(BAR_W, pts_in_window), col))
        add("")

    rule("─")

    # ── Totals ────────────────────────────────────────────────────────────────
    t_now = sum(snap_pw.values())
    t_peak = sum(snap_pk.values())
    t_avg = sum(state.rolling_avg(k) for k in ("CPU", "GPU", "ANE"))
    t_mj = sum(snap_en.values())

    add(
        rpad(
            C_WHITE
            + "\033[1m  Σ TOTAL"
            + R
            + C_WARN
            + "\033[1m"
            + f"  {t_now:>5} mW"
            + R
            + C_DIM
            + f"  pk {t_peak} mW"
            + R
            + C_DIM
            + f"  avg {t_avg:.0f} mW"
            + R
            + C_DIM
            + f"  session Σ {fmt_energy(t_mj)}"
            + R,
            W,
        )
    )
    rule("─")

    # ── Keybinds ──────────────────────────────────────────────────────────────
    def badge(k, d):
        return (
            bg(40, 45, 58) + C_WHITE + "\033[1m" + f" {k} " + R + C_DIM + f" {d}  " + R
        )

    kb = "  "
    for k, d in [
        ("q", "quit"),
        ("p", "pause"),
        ("r", "reset pk"),
        ("h", "history"),
        ("+ -", "interval"),
    ]:
        kb += badge(k, d)
    add(kb)
    add("")

    # Join and prepend cursor-home (no clear = no flicker)
    return "\033[H" + "\r\n".join(out) + "\033[J"


# ── Powermetrics ──────────────────────────────────────────────────────────────

POWER_PATS = {
    "CPU": re.compile(r"CPU Power:.*?(\d+)\s*mW", re.IGNORECASE),
    "GPU": re.compile(r"GPU Power:.*?(\d+)\s*mW", re.IGNORECASE),
    "ANE": re.compile(r"ANE Power:.*?(\d+)\s*mW", re.IGNORECASE),
}


def start_reader(state: State):
    # Prompt for sudo password upfront securely
    subprocess.run(["sudo", "-v"], check=True)

    proc = subprocess.Popen(
        [
            "sudo",
            "-S",
            "powermetrics",
            "--samplers",
            "cpu_power,gpu_power,ane_power",
            "-i",
            str(state.interval),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        universal_newlines=True,
        bufsize=1,  # Line buffered
    )

    def _run():
        # Iterate over readline instead of proc.stdout directly to bypass some buffering issues
        for line in iter(proc.stdout.readline, ""):
            if not line:
                break
            for key, pat in POWER_PATS.items():
                m = pat.match(line.strip())
                if m:
                    state.record(key, int(m.group(1)))

    threading.Thread(target=_run, daemon=True).start()
    return proc


# ── Keyboard ──────────────────────────────────────────────────────────────────


class KeyReader:
    def __enter__(self):
        self._fd = sys.stdin.fileno()
        self._old = termios.tcgetattr(self._fd)
        tty.setcbreak(self._fd)
        return self

    def __exit__(self, *_):
        termios.tcsetattr(self._fd, termios.TCSADRAIN, self._old)

    def read(self, timeout=0.05) -> Optional[str]:
        r, _, _ = select.select([sys.stdin], [], [], timeout)
        return sys.stdin.read(1) if r else None


# ── Main ──────────────────────────────────────────────────────────────────────


def monitor(interval_ms: int = 500):
    state = State(interval_ms)
    proc = start_reader(state)

    sys.stdout.write("\033[?25l\033[2J\033[H")
    sys.stdout.flush()

    def cleanup(*_):
        sys.stdout.write("\033[?25h" + R + "\n")
        sys.stdout.flush()
        proc.terminate()
        try:
            proc.wait(timeout=2)
        except Exception:
            proc.kill()
        sys.exit(0)

    signal.signal(signal.SIGTERM, cleanup)
    signal.signal(signal.SIGINT, cleanup)

    with KeyReader() as keys:
        while True:
            key = keys.read(timeout=0.08)
            if key:
                k = key.lower()
                if k in ("q", "\x03", "\x04"):
                    cleanup()
                elif k == "p":
                    state.paused = not state.paused
                elif k == "r":
                    state.reset_peaks()
                elif k == "h":
                    state.hist_idx += 1
                elif k == "+":
                    state.interval = max(100, state.interval - 100)
                elif k == "-":
                    state.interval = min(5000, state.interval + 100)

            sys.stdout.write(render(state))
            sys.stdout.flush()
            time.sleep(0.1)


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("-i", "--interval", type=int, default=500)
    args = ap.parse_args()
    monitor(interval_ms=args.interval)


if __name__ == "__main__":
    main()



import pytest
from energizer.autograd import Tensor
from energizer.coreml.tracer import Tracer, TraceData


def test_tracing_basic_mlp():
    with Tracer() as tracer:
        # Create trace-friendly Tensors
        x = Tensor(TraceData((32, 784), "float32"))
        w1 = Tensor(TraceData((784, 128), "float32"), requires_grad=True)
        b1 = Tensor(TraceData((128,), "float32"), requires_grad=True)
        w2 = Tensor(TraceData((128, 10), "float32"), requires_grad=True)
        b2 = Tensor(TraceData((10,), "float32"), requires_grad=True)

        # Manual Forward pass
        h = x @ w1
        h = h + b1
        h = h.maximum(0.0)
        out = h @ w2
        out = out + b2

    nodes = tracer.nodes

    # Check that ops were recorded
    assert len(nodes) == 5

    # 1. MatMul
    assert nodes[0].op == "MatMul"
    assert nodes[0].output_shape == (32, 128)

    # 2. Add
    assert nodes[1].op == "Add"
    assert nodes[1].output_shape == (32, 128)

    # 3. Maximum (ReLU)
    assert nodes[2].op == "Maximum"
    assert nodes[2].output_shape == (32, 128)

    # 4. MatMul
    assert nodes[3].op == "MatMul"
    assert nodes[3].output_shape == (32, 10)

    # 5. Add
    assert nodes[4].op == "Add"
    assert nodes[4].output_shape == (32, 10)


def test_tracing_shape_inference():
    with Tracer() as tracer:
        x = Tensor(TraceData((10, 20), "float32"))

        # SumAxis
        s = x.sum(axis=0)

        # Transpose
        t = x.T

        # Reshape
        r = x.reshape((5, -1))

        # Squeeze
        sq = r.reshape((5, 40, 1)).squeeze(axis=2)

    nodes = tracer.nodes
    assert len(nodes) == 5
    assert nodes[0].op == "SumAxis"
    assert nodes[0].output_shape == (20,)

    assert nodes[1].op == "Transpose"
    assert nodes[1].output_shape == (20, 10)

    assert nodes[2].op == "Reshape"
    assert nodes[2].output_shape == (5, 40)

    # squeeze actually calls reshape first, so nodes[3] is Reshape
    assert nodes[3].op == "Reshape"
    assert nodes[3].output_shape == (5, 40, 1)

    assert nodes[4].op == "Squeeze"
    assert nodes[4].output_shape == (5, 40)



from typing import Union
from energizer.tensor import Tensor
import energizer.autograd as autograd
import numpy as np


def max(tensor: Tensor, floor: Union[int, float] = 0) -> Tensor:
    return tensor.maximum(floor)


def as_strided(
    tensor: Tensor, shape: tuple, strides: tuple, storage_offset: int = 0
) -> "Tensor":
    return autograd.AsStrided.apply(tensor, shape, strides, storage_offset)


def trace(tensor: Tensor) -> Tensor:
    return autograd.Trace.apply(tensor)


def tanh(tensor: Tensor) -> Tensor:
    return autograd.Tanh.apply(tensor)


def softmax(tensor: Tensor) -> Tensor:
    return autograd.Softmax.apply(tensor)



import coremltools as ct
from coremltools.converters.mil import Builder as mb
from .tracer import Tracer, IRNode
from energizer.tensor import Tensor
import numpy as np


class Transpiler:
    def __init__(self):
        self._op_mapping = {
            "MatMul": self._translate_matmul,
            "Add": self._translate_add,
            "Maximum": self._translate_relu,
            "Transpose": self._translate_transpose,
            "Mul": self._translate_mul,
            "Sigmoid": self._translate_sigmoid,
            "ReLU": self._translate_relu,
        }

    def transpile(self, model, example_inputs, output_name="model.mlpackage"):
        """
        Transpile a traced Energizer model into a CoreML package.
        """
        with Tracer() as tracer:
            nodes = tracer.trace(model, example_inputs)

        # Define the input specification based on the example input shape
        input_shape = example_inputs[0].shape
        input_spec = [mb.TensorSpec(shape=input_shape)]

        @mb.program(input_specs=input_spec)
        def prog(x):
            # Maps energizer object ids to CoreML MIL variables
            self.env = {id(example_inputs[0]): x}
            last_out = x

            for node in nodes:
                op_translator = self._op_mapping.get(node.op)
                if not op_translator:
                    raise NotImplementedError(
                        f"Transpilation for op '{node.op}' is not supported yet."
                    )

                last_out = op_translator(node)
                self.env[id(node)] = last_out

            return last_out

        print(f"Converting traced program to CoreML...")
        try:
            coreml_model = ct.convert(prog)
            coreml_model.save(output_name)
            print(f"Model successfully saved to {output_name}")
            return coreml_model
        except Exception as e:
            print(
                f"CoreML Conversion Failed. Note: This could be due to Python 3.14 incompatibility with coremltools."
            )
            raise e

    def _get_input_var(self, inp):
        if isinstance(inp, IRNode):
            return self.env[id(inp)]
        elif isinstance(inp, Tensor):
            return mb.const(val=inp.numpy())
        else:
            return mb.const(val=inp)

    def _translate_matmul(self, node):
        x = self._get_input_var(node.inputs[0])
        w = self._get_input_var(node.inputs[1])
        return mb.matmul(x=x, y=w)

    def _translate_add(self, node):
        a = self._get_input_var(node.inputs[0])
        b = self._get_input_var(node.inputs[1])
        return mb.add(x=a, y=b)

    def _translate_relu(self, node):
        x = self._get_input_var(node.inputs[0])
        return mb.relu(x=x)

    def _translate_transpose(self, node):
        x = self._get_input_var(node.inputs[0])
        # Find dimension logic based on shape
        inp = node.inputs[0]
        rank = (
            len(inp.shape) if hasattr(inp, "shape") else 1
        )  # Fallback to 1 if scalar/unknown
        perm = list(reversed(range(max(1, rank))))
        return mb.transpose(x=x, perm=perm)

    def _translate_mul(self, node):
        x = self._get_input_var(node.inputs[0])
        y = self._get_input_var(node.inputs[1])
        return mb.multiply(x=x, y=y)

    def _translate_sigmoid(self, node):
        x = self._get_input_var(node.inputs[0])
        return mb.sigmoid(x=x)

    def _translate_relu(self, node):
        x = self._get_input_var(node.inputs[0])
        return mb.relu(x=x)


def compile_to_coreml(model, example_input, output_path="model.mlpackage"):
    transpiler = Transpiler()
    if not isinstance(example_input, tuple):
        example_input = (example_input,)
    return transpiler.transpile(model, example_input, output_path)



from typing import Union
from energizer.neural_network import Module
import numpy as np

from energizer.tensor import Tensor
from energizer.functionnal import max


class ReLU(Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        return max(x)

    def extra_repr(self):
        return ""


class LeakyReLU(Module):
    def __init__(self, negative_slope: float = 0.01, inplace: bool = False):
        super().__init__()
        self.negative_slope = negative_slope
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        return max(x) + self.negative_slope * (x - max(x))

    def extra_repr(self):
        return ""



from energizer.neural_network import Module
from energizer.tensor import Tensor
import energizer.autograd as autograd
import numpy as np

try:
    import mlx.core as mx
except ImportError:
    mx = None


class MSELoss(Module):
    def __init__(
        self, size_average: bool = None, reduce: bool = None, reduction: str = "mean"
    ):
        super().__init__()
        self.reduction = reduction

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        diff = (input - target) ** 2
        if self.reduction == "sum":
            return diff.sum()
        return diff.mean()


class CrossEntropyFn(autograd.Function):
    @staticmethod
    def forward(ctx, logits_data, target_data, reduction):
        ctx.save_for_backward(logits_data, target_data, reduction)

        logits_np = np.array(logits_data).astype(np.float32)
        target_np = np.array(target_data)

        B = logits_np.shape[0]

        shifted = logits_np - logits_np.max(axis=1, keepdims=True)
        log_probs = shifted - np.log(np.exp(shifted).sum(axis=1, keepdims=True))

        if target_np.ndim == 1:
            nll = -log_probs[np.arange(B), target_np.astype(int)]
        else:
            nll = -(target_np.astype(np.float32) * log_probs).sum(axis=1)

        if reduction == "sum":
            loss_val = float(nll.sum())
        elif reduction == "mean":
            loss_val = float(nll.mean())
        else:
            loss_val = nll

        if isinstance(loss_val, float):
            return np.array(loss_val, dtype=np.float32)
        return loss_val.astype(np.float32)

    @staticmethod
    def backward(ctx, grad):
        logits_data, target_data, reduction = ctx.saved_tensors

        logits_np = np.array(logits_data).astype(np.float32)
        target_np = np.array(target_data)

        B = logits_np.shape[0]
        shifted = logits_np - logits_np.max(axis=1, keepdims=True)
        softmax_probs = np.exp(shifted)
        softmax_probs /= softmax_probs.sum(axis=1, keepdims=True)

        grad_logits = softmax_probs.copy()
        if target_np.ndim == 1:
            grad_logits[np.arange(B), target_np.astype(int)] -= 1.0
        else:
            grad_logits -= target_np.astype(np.float32)

        if reduction == "mean":
            grad_logits /= B

        upstream = float(np.array(grad).flat[0])
        grad_logits = (grad_logits * upstream).astype(np.float32)

        return (grad_logits, None, None)


class CrossEntropyLoss(Module):
    def __init__(
        self, size_average: bool = None, reduce: bool = None, reduction: str = "mean"
    ):
        super().__init__()
        self.reduction = reduction

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        return CrossEntropyFn.apply(input, target, self.reduction)



"""
Energizer — A lightweight deep learning library for Apple's Neural Engine.

Energizer provides a PyTorch-like API for building, training, and running
neural networks with first-class support for Apple Silicon via the MLX backend.
It can also run on CPU using NumPy as a fallback, making it suitable for
prototyping and experimentation on any platform.

The library implements automatic differentiation, a modular layer system,
and GPU-accelerated tensor operations through a familiar interface.

Core Features:
    - **Tensor**: Multi-backend tensor with autograd support (CPU via NumPy, GPU via MLX).
    - **Module**: Base class for all neural network layers with parameter management,
      serialization (``save``/``load``), device transfer (``to``), and train/eval modes.
    - **Optimizer**: Base optimizer with ``step()``, ``zero_grad()``, and state management.

Tensor Operations:
    - Creation: ``tensor()``, ``Tensor.randn()``, ``Tensor.zeros()``, ``Tensor.ones()``
    - Arithmetic: ``+``, ``-``, ``*``, ``/``, ``@`` (matmul), ``**`` (power)
    - Reduction: ``sum()``, ``mean()``
    - Shape: ``reshape()``, ``view()``, ``T`` (transpose)
    - Device: ``cpu()``, ``gpu()``, ``to(device)``
    - Autograd: ``backward()`` for automatic gradient computation

Functional API (``energizer.functionnal``):
    - ``max(tensor, floor)`` — Element-wise maximum with a floor value.
    - ``as_strided(tensor, shape, strides)`` — Create a strided view of a tensor.
    - ``trace(tensor)`` — Compute the trace of a 2D matrix.

Optimizers:
    - ``SGD``  — Stochastic Gradient Descent with momentum, weight decay, and Nesterov support.
    - ``Adam`` — Adaptive Moment Estimation with bias correction and AMSGrad variant.

Layers:
    - **Linear**: ``Linear(in_features, out_features, bias=True)``
      Fully connected layer: ``y = x @ W^T + b``.

    - **Convolutional**:
        - ``Conv1d(in_channels, out_channels, kernel_size, stride, padding)``
        - ``Conv2d(in_channels, out_channels, kernel_size, stride, padding)``
        - ``ConvTranspose2d(in_channels, out_channels, kernel_size, ...)``
          Transposed (deconvolution) layer for upsampling.

    - **Activation Functions**:
        - ``ReLU()``           — Rectified Linear Unit: ``max(0, x)``.
        - ``LeakyReLU(slope)`` — Leaky variant: ``max(0, x) + slope * min(0, x)``.
        - ``Sigmoid()``        — Logistic sigmoid: ``1 / (1 + exp(-x))``.

    - **Normalization**:
        - ``BatchNorm1d(num_features)`` — Batch normalization over 2D/3D input.
        - ``BatchNorm2d(num_features)`` — Batch normalization over 4D input.

    - **Pooling**:
        - ``MaxPool2d(kernel_size, stride, padding)``
        - ``AvgPool2d(kernel_size, stride, padding)``

    - **Regularization**:
        - ``Dropout(p=0.5)`` — Randomly zeroes elements during training.

    - **Shape Manipulation**:
        - ``Flatten(start_dim, end_dim)`` — Flattens contiguous dimensions.
        - ``Reshape(shape)``              — Reshapes tensor to target shape.
        - ``Trim(start, end)``            — Trims spatial dimensions of 4D tensors.

    - **Composite / Containers**:
        - ``Sequential(*layers)``      — Chains layers in order.
        - ``ModuleList(modules)``      — List-like container for modules (no ``forward``).
        - ``ResidualBlock(channels)``   — Conv → BN → ReLU → Conv → BN + skip connection.
        - ``BottleneckBlock(in_ch, out_ch)`` — Bottleneck residual block (expansion=4).
        - ``AutoEncoder(device)``       — Pre-configured convolutional autoencoder.

Loss Functions:
    - ``MSELoss(reduction='mean')``          — Mean Squared Error loss.
    - ``CrossEntropyLoss(reduction='mean')`` — Cross-Entropy loss for classification.

Example::

    import energizer

    # Create a simple model
    model = energizer.Sequential(
        energizer.Linear(784, 128),
        energizer.ReLU(),
        energizer.Linear(128, 10),
    )

    # Move to GPU (Apple Neural Engine via MLX)
    model.to('gpu')

    # Forward pass
    x = energizer.Tensor.randn(32, 784, device='gpu')
    output = model(x)

    # Compute loss and backpropagate
    loss_fn = energizer.MSELoss()
    target = energizer.Tensor.zeros(32, 10, device='gpu')
    loss = loss_fn(output, target)
    loss.backward()

    # Optimize
    optimizer = energizer.Adam(model.parameters(), lr=0.001)
    optimizer.step()
    optimizer.zero_grad()
"""

__version__ = "0.1.6.1"
__author__ = "Florian GRIMA"
__name__ = "energizer"
__description__ = "A lightweight deep learning library for Apple's Neural Engine."
__url__ = "https://github.com/fgrimaepitech/energizer"
__license__ = "MIT"
__copyright__ = "Copyright 2026 Florian GRIMA"
__maintainer__ = "Florian GRIMA"
__status__ = "Development"


# ---------------------------------------------------------------------------
# Context Managers
# ---------------------------------------------------------------------------
class _NoGrad:
    """
    Simple context manager to disable gradient tracking.

    This is a lightweight analogue of torch.no_grad(). It temporarily sets
    `requires_grad` to False on all existing tensors that currently require
    gradients, and restores their original flags on exit.
    """

    def __enter__(self):
        # Track tensors whose requires_grad flag we flip.
        self._modified = []
        # Import here to avoid circular imports at module import time.
        try:
            import gc
        except ImportError:
            return self

        for obj in gc.get_objects():
            try:
                if isinstance(obj, Tensor) and getattr(obj, "requires_grad", False):
                    self._modified.append(obj)
                    obj.requires_grad = False
            except ReferenceError:
                continue
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for t in getattr(self, "_modified", []):
            t.requires_grad = True
        return False


def no_grad():
    """
    Context manager that disables autograd on existing tensors.

    Usage:

        import energizer as ez

        with ez.no_grad():
            out = model(x)
    """

    return _NoGrad()


# ---------------------------------------------------------------------------
# Core
# ---------------------------------------------------------------------------
from .tensor import tensor, Tensor
from .neural_network import Module, Optimizer

# ---------------------------------------------------------------------------
# Functional API
# ---------------------------------------------------------------------------
from .functionnal import max, as_strided, trace

# ---------------------------------------------------------------------------
# Optimizers
# ---------------------------------------------------------------------------
from .optimizer.sgd import SGD
from .optimizer.adam import Adam

# ---------------------------------------------------------------------------
# Layers — Linear & Convolutional
# ---------------------------------------------------------------------------
from .layers.linear import Linear
from .layers.conv import ConvNd, Conv1d, Conv2d
from .layers.convtranspose import ConvTranspose2d

# ---------------------------------------------------------------------------
# Layers — Activation Functions
# ---------------------------------------------------------------------------
from .layers.relu import ReLU, LeakyReLU
from .layers.sigmoid import Sigmoid
from .layers.gelu import GELU
from .layers.tanh import Tanh

# ---------------------------------------------------------------------------
# Layers — Normalization
# ---------------------------------------------------------------------------
from .layers.batch_norm import BatchNorm1d, BatchNorm2d
from .layers.layer_norm import LayerNorm

# ---------------------------------------------------------------------------
# Layers — Pooling
# ---------------------------------------------------------------------------
from .layers.pool import MaxPool2d, AvgPool2d

# ---------------------------------------------------------------------------
# Layers — Regularization
# ---------------------------------------------------------------------------
from .layers.dropout import Dropout

# ---------------------------------------------------------------------------
# Layers — Shape Manipulation
# ---------------------------------------------------------------------------
from .layers.flatten import Flatten
from .layers.reshape import Reshape
from .layers.trim import Trim

# ---------------------------------------------------------------------------
# Layers — Composite / Containers
# ---------------------------------------------------------------------------
from .layers.sequential import Sequential
from .layers.modulelist import ModuleList
from .layers.residual import ResidualBlock, BottleneckBlock
from .layers.autoencoder import AutoEncoder

# ---------------------------------------------------------------------------
# Layers — Transformer
# ---------------------------------------------------------------------------
from .layers.transformer import TransformerEncoderLayer, TransformerEncoder

# ---------------------------------------------------------------------------
# Loss Functions
# ---------------------------------------------------------------------------
from .layers.loss import MSELoss, CrossEntropyLoss

# ---------------------------------------------------------------------------
# Layers — Embedding
# ---------------------------------------------------------------------------
from .layers.embedding import Embedding

# ---------------------------------------------------------------------------
# Initializers
# ---------------------------------------------------------------------------
from .layers.init import zeros_, orthogonal_



from __future__ import annotations
from typing import Any, Callable, List


class Function:
    def __init__(self, function: Callable, tensors: List[Any], **kwargs: Any):
        self.function = function
        self.tensors = tensors
        self.kwargs = kwargs

    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:
        pass

    def backward(ctx: Any, grad_outputs: Any) -> Any:
        return ctx.function(ctx.tensors, grad_outputs, **ctx.kwargs)



from typing import Any


class IRNode:
    def __init__(self, op: str, inputs: list, output_shape: tuple, output_dtype: str):
        self.op = op
        self.inputs = inputs
        self.output_shape = output_shape
        self.output_dtype = output_dtype

    def __repr__(self):
        return f"IRNode(op='{self.op}', inputs={len(self.inputs)}, shape={self.output_shape}, dtype='{self.output_dtype}')"


class TraceData:
    def __init__(self, shape: tuple, dtype: str = "float32"):
        if isinstance(shape, int):
            shape = (shape,)
        self.shape = tuple(shape) if shape else ()
        self.dtype = dtype

    @property
    def size(self):
        import numpy as np

        return int(np.prod(self.shape)) if self.shape else 1


class Tracer:
    _active = None

    def __init__(self):
        self.nodes = []
        self._tensor_to_node: dict[int, IRNode] = (
            {}
        )  # id(Tensor) → IRNode that produced it

    def __enter__(self):
        Tracer._active = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        Tracer._active = None

    @classmethod
    def is_tracing(cls):
        return cls._active is not None

    @classmethod
    def get(cls):
        return cls._active

    def trace(self, model, example_inputs):
        self.nodes = []
        self._tensor_to_node = {}
        with self:
            model(*example_inputs)
        return self.nodes

    def record(self, op_name: str, input_tensors: list, result) -> IRNode:
        """
        Record one op into the IR graph.
        Resolves each input: if it was produced by a previous op, store the IRNode
        reference. Otherwise store the raw Tensor (weight / constant).
        """
        resolved_inputs = []
        for t in input_tensors:
            if id(t) in self._tensor_to_node:
                resolved_inputs.append(self._tensor_to_node[id(t)])
            else:
                resolved_inputs.append(t)  # weight or constant

        node = IRNode(
            op=op_name,
            inputs=resolved_inputs,
            output_shape=result.shape,
            output_dtype="float32",
        )
        self.nodes.append(node)
        self._tensor_to_node[id(result)] = node
        return node

    @classmethod
    def infer_shape(cls, op: str, *args):
        import numpy as np

        shapes = []
        for x in args:
            if isinstance(x, TraceData):
                shapes.append(x.shape)
            elif hasattr(x, "shape"):
                shapes.append(x.shape)
            elif isinstance(x, (int, float, complex, bool)):
                shapes.append(())
            elif isinstance(x, (list, tuple)) and op not in (
                "Reshape",
                "AsStrided",
                "SumAxis",
                "Squeeze",
            ):
                shapes.append(np.array(x).shape)

        if op in ("Add", "Sub", "Mul", "Div", "Minimum", "Maximum", "Pow"):
            return np.broadcast_shapes(shapes[0], shapes[1])

        if op in ("Neg", "Exp", "Log", "Clamp", "Sigmoid", "Tanh", "GELU", "Softmax"):
            return shapes[0]

        if op == "MatMul":
            shape_a, shape_b = shapes[0], shapes[1]
            if len(shape_a) == 0 or len(shape_b) == 0:
                raise ValueError("MatMul requires at least 1D tensors")
            if len(shape_a) == 1 and len(shape_b) == 1:
                return ()
            elif len(shape_a) == 1:
                return shape_b[:-2] + (shape_b[-1],)
            elif len(shape_b) == 1:
                return shape_a[:-1]
            else:
                batch_shape = np.broadcast_shapes(shape_a[:-2], shape_b[:-2])
                return batch_shape + (shape_a[-2], shape_b[-1])

        if op in ("Sum", "Mean"):
            return ()

        if op == "SumAxis":
            axis = args[1]
            shape = list(shapes[0])
            for ax in sorted(
                [axis] if isinstance(axis, int) else list(axis), reverse=True
            ):
                shape.pop(ax)
            return tuple(shape)

        if op == "Squeeze":
            axis = args[1] if len(args) > 1 else None
            shape = list(shapes[0])
            if axis is None:
                return tuple(s for s in shape if s != 1)
            for ax in sorted(
                [axis] if isinstance(axis, int) else list(axis), reverse=True
            ):
                if shape[ax] == 1:
                    shape.pop(ax)
            return tuple(shape)

        if op == "Reshape":
            new_shape = list(args[1])
            if -1 in new_shape:
                idx = new_shape.index(-1)
                total = int(np.prod(shapes[0])) if shapes[0] else 1
                known = int(np.prod([s for s in new_shape if s != -1]))
                new_shape[idx] = total // known
            return tuple(new_shape)

        if op == "Transpose":
            return tuple(reversed(shapes[0]))

        if op == "GetItem":
            dummy = np.empty(shapes[0], dtype=np.float32)[args[1]]
            return dummy.shape

        if op == "Trace":
            return shapes[0][:-2] if len(shapes[0]) >= 2 else ()

        if op == "AsStrided":
            return args[1]

        return shapes[0] if shapes else ()



from energizer.neural_network import Module
from energizer.tensor import Tensor
from energizer.function import Function
import energizer.derivatives as dv
import numpy as np


class Dropout(Module):
    def __init__(self, p: float = 0.5):
        super().__init__()
        if p < 0 or p > 1:
            raise ValueError("p must be between 0 and 1")
        self.p = p
        self.training = True

    def forward(self, x: Tensor) -> Tensor:
        if not self.training or self.p == 0:
            return x

        scale = 1.0 / (1.0 - self.p)
        # Always generate mask in numpy; convert to the right backend below
        mask = (np.random.rand(*x.data.shape) > self.p).astype(np.float32) * scale

        try:
            import mlx.core as mx

            if isinstance(x.data, mx.array):
                out = x.data * mx.array(mask)
            else:
                out = x.data * mask
        except ImportError:
            out = x.data * mask

        return Tensor(
            out,
            requires_grad=x.requires_grad,
            device=x.device,
        )

    def eval(self):
        self.training = False

    def train(self, mode: bool = True):
        self.training = mode



from energizer.neural_network import Module
from energizer.tensor import Tensor
import energizer.autograd as autograd


class Tanh(Module):
    def __init__(self, device: str = "cpu"):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        return autograd.Tanh.apply(x)



from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np


class Trim(Module):
    def __init__(self, start: int, end: int):
        super().__init__()
        self.start = start
        self.end = end

    def forward(self, x: Tensor) -> Tensor:
        h_start = self.start
        h_end = -self.end if self.end != 0 else None
        w_start = self.start
        w_end = -self.end if self.end != 0 else None
        trimmed = x.data[:, :, h_start:h_end, w_start:w_end]
        return Tensor(trimmed, requires_grad=x.requires_grad)



from energizer.neural_network import Module
import energizer
from energizer.tensor import Tensor


class AutoEncoder(Module):
    def __init__(self, device: str = "cpu"):
        super().__init__()
        self.device = device
        self.encoder = energizer.Sequential(
            energizer.Conv2d(
                in_channels=1,
                out_channels=32,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=1,
            ),
            energizer.LeakyReLU(0.01),
            energizer.Conv2d(
                in_channels=32,
                out_channels=64,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=1,
            ),
            energizer.LeakyReLU(0.01),
            energizer.Conv2d(
                in_channels=64,
                out_channels=64,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=1,
            ),
            energizer.LeakyReLU(0.01),
            energizer.Conv2d(
                in_channels=64,
                out_channels=64,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=1,
            ),
            energizer.Flatten(),
            energizer.Linear(3136, 2),
        )

        self.decoder = energizer.Sequential(
            energizer.Linear(2, 3136),
            energizer.Reshape((-1, 64, 7, 7)),
            energizer.ConvTranspose2d(
                64, 64, stride=(1, 1), kernel_size=(3, 3), padding=1
            ),
            energizer.LeakyReLU(0.01),
            energizer.ConvTranspose2d(
                64, 64, stride=(2, 2), kernel_size=(3, 3), padding=1
            ),
            energizer.LeakyReLU(0.01),
            energizer.ConvTranspose2d(
                64, 32, stride=(2, 2), kernel_size=(3, 3), padding=0
            ),
            energizer.LeakyReLU(0.01),
            energizer.ConvTranspose2d(
                32, 1, stride=(1, 1), kernel_size=(3, 3), padding=1, output_padding=1
            ),
            energizer.Sigmoid(),
        )

        self.add_module("encoder", self.encoder)
        self.add_module("decoder", self.decoder)

    def forward(self, x: Tensor) -> Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def train(self, mode: bool = True):
        self.encoder.train(mode)
        self.decoder.train(mode)
        return self

    def to(self, device: str):
        self.device = device
        self.encoder.to(device)
        self.decoder.to(device)
        return self



from __future__ import annotations
from typing import Any
import numpy as np
import mlx.core as mx


def _get_data_ref(x: Any):
    return x.data if hasattr(x, "data") else x


def zeros_(tensor: Any) -> Any:
    data = _get_data_ref(tensor)
    if mx is not None and isinstance(data, mx.array):
        z = mx.zeros_like(data)
    else:
        z = np.zeros_like(np.array(data))
    if hasattr(tensor, "data"):
        tensor.data = z
        return tensor
    return z


def orthogonal_(tensor: Any, gain: float = 1.0, seed: int | None = None) -> Any:
    data = _get_data_ref(tensor)
    shape = getattr(data, "shape", None)
    if shape is None or len(shape) < 2:
        raise ValueError("orthogonal_ expects a tensor with at least 2 dimensions")
    if seed is not None:
        rng = np.random.default_rng(seed)
        randn = rng.standard_normal
    else:
        randn = np.random.randn
    rows = int(shape[0])
    cols = int(np.prod(shape[1:]))
    flat_shape = (rows, cols)
    a = randn(*flat_shape).astype(np.float32)
    transposed = False
    if rows < cols:
        a = a.T
        transposed = True
    q, r = np.linalg.qr(a)
    d = np.diag(r)
    q *= np.sign(d + 1e-12)
    if transposed:
        q = q.T
    q = (q.reshape(shape) * float(gain)).astype(np.float32)
    if mx is not None and isinstance(data, mx.array):
        out = mx.array(q)
    else:
        out = q
    if hasattr(tensor, "data"):
        tensor.data = out
        return tensor
    return out



from tkinter import N
import energizer
from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np


class ResidualBlock(Module):
    def __init__(self, channels: int):
        super().__init__()
        self.conv1 = energizer.Conv2d(channels, channels, kernel_size=3, padding=1)
        self.bn1 = energizer.BatchNorm2d(channels)
        self.conv2 = energizer.Conv2d(channels, channels, kernel_size=3, padding=1)
        self.bn2 = energizer.BatchNorm2d(channels)

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = energizer.ReLU()(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += identity
        out = energizer.ReLU()(out)
        return out


class BottleneckBlock(Module):
    def __init__(self, in_channels: int, out_channels: int, stride: int = 1):
        super().__init__()
        self.expansion = 4
        reduce_channels = out_channels // self.expansion

        self.conv1 = energizer.Conv2d(
            in_channels, reduce_channels, kernel_size=1, bias=False
        )
        self.bn1 = energizer.BatchNorm2d(reduce_channels)

        self.conv2 = energizer.Conv2d(
            reduce_channels,
            reduce_channels,
            kernel_size=3,
            padding=1,
            stride=stride,
            bias=False,
        )
        self.bn2 = energizer.BatchNorm2d(reduce_channels)

        self.conv3 = energizer.Conv2d(
            reduce_channels, out_channels, kernel_size=1, bias=False
        )
        self.bn3 = energizer.BatchNorm2d(out_channels)

        if stride != 1 or in_channels != out_channels:
            self.shortcut = energizer.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=stride
            )
            self.shortcut_bn = energizer.BatchNorm2d(out_channels)
        else:
            self.shortcut = None

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = energizer.ReLU()(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = energizer.ReLU()(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.shortcut is not None:
            identity = self.shortcut(identity)
            identity = self.shortcut_bn(identity)

        out += identity
        out = energizer.ReLU()(out)

        return out



"""
Tests comparing energizer Linear layer with PyTorch Linear layer.
Tests forward pass, backward pass, and various configurations.
"""

import unittest
import numpy as np
import torch
import torch.nn as nn
import sys
import os

# Add parent directory to path to import energizer
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import energizer


class TestLinear(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures"""
        self.rtol = 1e-4  # Relative tolerance for floating point comparison
        self.atol = (
            1e-6  # Absolute tolerance for floating point comparison (float32 precision)
        )
        np.random.seed(42)  # Set seed for reproducibility
        torch.manual_seed(42)

    def assert_tensors_close(self, my_tensor, torch_tensor, msg=""):
        """Helper to compare energizer tensor with PyTorch tensor"""
        # Convert PyTorch tensor to numpy if needed
        if isinstance(torch_tensor, torch.Tensor):
            torch_data = torch_tensor.detach().numpy()
        else:
            torch_data = np.array(torch_tensor)

        # Compare data
        np.testing.assert_allclose(
            my_tensor.data,
            torch_data,
            rtol=self.rtol,
            atol=self.atol,
            err_msg=f"Tensor data mismatch. {msg}",
        )

    def set_same_weights(self, my_layer, torch_layer):
        """Set the same weights and biases for both layers"""
        # Get PyTorch weights and biases
        weight_data = torch_layer.weight.data.detach().numpy()
        bias_data = (
            torch_layer.bias.data.detach().numpy()
            if torch_layer.bias is not None
            else None
        )

        # Set energizer weights and biases
        my_layer.weight.data = weight_data
        if bias_data is not None:
            my_layer.bias.data = bias_data

    # ========== FORWARD PASS TESTS ==========

    def test_linear_forward_basic(self):
        """Test basic forward pass: (128, 20) -> (128, 30)"""
        in_features, out_features = 20, 30
        batch_size = 128

        # Create layers
        my_layer = energizer.Linear(in_features, out_features)
        torch_layer = nn.Linear(in_features, out_features)

        # Set same weights
        self.set_same_weights(my_layer, torch_layer)

        # Create input
        input_data = np.random.randn(batch_size, in_features).astype(np.float32)
        my_input = energizer.tensor(input_data)
        torch_input = torch.tensor(input_data)

        # Forward pass
        my_output = my_layer(my_input)
        torch_output = torch_layer(torch_input)

        # Compare outputs
        self.assert_tensors_close(my_output, torch_output, "Linear forward pass")

    def test_linear_forward_bias_false(self):
        """Test forward pass with bias=False: (128, 20) -> (128, 30)"""
        in_features, out_features = 20, 30
        batch_size = 128

        # Create layers
        my_layer = energizer.Linear(in_features, out_features, bias=False)
        torch_layer = nn.Linear(in_features, out_features, bias=False)

        # Set same weights
        self.set_same_weights(my_layer, torch_layer)

        # Create input
        input_data = np.random.randn(batch_size, in_features).astype(np.float32)
        my_input = energizer.tensor(input_data)
        torch_input = torch.tensor(input_data)

        # Forward pass
        my_output = my_layer(my_input)
        torch_output = torch_layer(torch_input)

        # Compare outputs
        self.assert_tensors_close(
            my_output, torch_output, "Linear forward pass with bias=False"
        )


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Union
from energizer.neural_network import Module
from energizer.tensor import Tensor
import energizer
import mlx.core as mx
import numpy as np


class MultiheadAttention(Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout: float = 0.1,
        bias: bool = True,
        device: str = "cpu",
    ):
        super().__init__(device=device)
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout_p = dropout
        self.bias = bias

        self.head_dim = embed_dim // num_heads
        self.q_proj = energizer.Linear(embed_dim, embed_dim, bias=bias, device=device)
        self.k_proj = energizer.Linear(embed_dim, embed_dim, bias=bias, device=device)
        self.v_proj = energizer.Linear(embed_dim, embed_dim, bias=bias, device=device)
        self.out_proj = energizer.Linear(embed_dim, embed_dim, bias=bias, device=device)

        self.dropout_layer = energizer.Dropout(dropout)

    def _reshape_for_attention(self, x: Tensor) -> Tensor:
        batch_size, seq_len, embed_dim = x.shape

        if self.device == "gpu":
            x_reshaped = mx.reshape(
                x.data, (batch_size, seq_len, self.num_heads, self.head_dim)
            )
            x_reshaped = mx.transpose(x_reshaped, (0, 2, 1, 3))
        else:
            x_reshaped = x.data.reshape(
                batch_size, seq_len, self.num_heads, self.head_dim
            )
            x_reshaped = x_reshaped.transpose(0, 2, 1, 3)

        return Tensor(x_reshaped, requires_grad=x.requires_grad, device=x.device)

    def _scaled_dot_product_attention(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        attn_mask: Tensor = None,
        key_padding_mask: Tensor = None,
    ) -> Tensor:
        if self.device == "gpu":
            scores = mx.matmul(q.data, mx.transpose(k.data, (0, 1, 3, 2)))
        else:
            scores = np.matmul(q.data, k.data.transpose(0, 1, 3, 2))

        scores = scores / (self.head_dim**0.5)

        if attn_mask is not None:
            scores = scores + attn_mask.data

        if key_padding_mask is not None:
            if self.device == "gpu":
                mask = key_padding_mask.data[:, None, None, :]
                scores = mx.where(mask, mx.full_like(scores, -float("inf")), scores)
            else:
                mask = key_padding_mask.data[:, None, None, :]
                scores = np.where(mask, -np.inf, scores)

        if self.device == "gpu":
            attn_weights = mx.softmax(scores, axis=-1)
            attn_weights = self.dropout_layer(Tensor(attn_weights, device="gpu")).data
            if not isinstance(attn_weights, mx.array):
                attn_weights = mx.array(attn_weights)
        else:
            exp_scores = np.exp(scores - np.max(scores, axis=-1, keepdims=True))
            attn_weights = exp_scores / np.sum(exp_scores, axis=-1, keepdims=True)
            attn_weights = self.dropout_layer(Tensor(attn_weights)).data

        if self.device == "gpu":
            output = mx.matmul(attn_weights, v.data)
        else:
            output = np.matmul(attn_weights, v.data)

        batch_size, num_heads, seq_len, head_dim = output.shape

        if self.device == "gpu":
            output = mx.transpose(output, (0, 2, 1, 3))
            output = mx.reshape(output, (batch_size, seq_len, self.embed_dim))
        else:
            output = output.transpose(0, 2, 1, 3)
            output = output.reshape(batch_size, seq_len, self.embed_dim)

        return Tensor(output, requires_grad=q.requires_grad, device=q.device)

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        attn_mask: Tensor = None,
        key_padding_mask: Tensor = None,
    ) -> Tensor:
        q = self.q_proj(query)
        k = self.k_proj(key)
        v = self.v_proj(value)

        q = self._reshape_for_attention(q)
        k = self._reshape_for_attention(k)
        v = self._reshape_for_attention(v)

        attn_output = self._scaled_dot_product_attention(
            q, k, v, attn_mask, key_padding_mask
        )

        output = self.out_proj(attn_output)

        return output


class TransformerEncoderLayer(Module):
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable, Tensor] = "relu",
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        device: str = "cpu",
    ):
        super().__init__(device=device)
        self.d_model = d_model
        self.nhead = nhead
        self.batch_first = batch_first
        self.norm_first = norm_first

        self.dropout_p = dropout
        self.self_attn = MultiheadAttention(
            d_model, nhead, dropout=dropout, bias=bias, device=device
        )

        self.linear1 = energizer.Linear(
            d_model, dim_feedforward, bias=bias, device=device
        )
        self.linear2 = energizer.Linear(
            dim_feedforward, d_model, bias=bias, device=device
        )

        self.dropout = energizer.Dropout(dropout)
        self.dropout1 = energizer.Dropout(dropout)
        self.dropout2 = energizer.Dropout(dropout)

        self.norm1 = energizer.LayerNorm(d_model, eps=layer_norm_eps, device=device)
        self.norm2 = energizer.LayerNorm(d_model, eps=layer_norm_eps, device=device)

        if activation == "relu":
            self.activation = energizer.ReLU()
        elif activation == "gelu":
            self.activation = energizer.GELU()

    def _sa_block(
        self, x: Tensor, attn_mask: Tensor = None, key_padding_mask: Tensor = None
    ) -> Tensor:
        x = self.self_attn(
            x, x, x, attn_mask=attn_mask, key_padding_mask=key_padding_mask
        )
        return self.dropout1(x)

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear1(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.linear2(x)
        return self.dropout2(x)

    def forward(
        self, src: Tensor, src_mask: Tensor = None, src_key_padding_mask: Tensor = None
    ) -> Tensor:
        if not self.batch_first:
            src = src.transpose(0, 1)

        if self.norm_first:
            x = src + self._sa_block(self.norm1(src), src_mask, src_key_padding_mask)
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(src + self._sa_block(src, src_mask, src_key_padding_mask))
            x = self.norm2(x + self._ff_block(x))

        if not self.batch_first:
            x = x.transpose(0, 1)

        return x

    def extra_repr(self):
        return f"d_model={self.d_model}, nhead={self.nhead}, dim_feedforward={self.dim_feedforward}, dropout={self.dropout}, activation={self.activation.__class__.__name__}, layer_norm_eps={self.layer_norm_eps}, batch_first={self.batch_first}, norm_first={self.norm_first}, bias={self.bias}"


class TransformerEncoder(Module):
    def __init__(
        self,
        encoder_layer: TransformerEncoderLayer,
        num_layers: int,
        norm: Module = None,
        device: str = "cpu",
    ):
        super().__init__(device=device)
        self.layers = energizer.ModuleList(
            [
                encoder_layer.__class__(
                    encoder_layer.d_model,
                    encoder_layer.nhead,
                    encoder_layer.linear1.out_features,
                    dropout=encoder_layer.dropout_p,
                    batch_first=encoder_layer.batch_first,
                    norm_first=encoder_layer.norm_first,
                    device=device,
                )
                for _ in range(num_layers)
            ]
        )

        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self, src: Tensor, src_mask: Tensor = None, src_key_padding_mask: Tensor = None
    ) -> Tensor:
        output = src

        for layer in self.layers:
            output = layer(output, src_mask, src_key_padding_mask=src_key_padding_mask)

        if self.norm is not None:
            output = self.norm(output)

        return output



from energizer import tensor
from energizer.neural_network import Module, Parameter
import numpy as np
import mlx.core as mx


class Linear(Module):
    def __init__(self, in_features, out_features, bias=True, device: str = "cpu"):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias
        self.device = device
        if device == "gpu":
            self.weight = Parameter(
                mx.random.normal(shape=(out_features, in_features)) * 0.01,
                requires_grad=True,
                device=device,
            )
        else:
            self.weight = Parameter(
                np.random.randn(out_features, in_features) * 0.01,
                requires_grad=True,
                device=device,
            )
        if bias:
            if device == "gpu":
                self.bias = Parameter(
                    mx.zeros(out_features), requires_grad=True, device=device
                )
            else:
                self.bias = Parameter(
                    np.zeros(out_features), requires_grad=True, device=device
                )
        else:
            self.bias = None

    def forward(self, x):
        result = x @ self.weight.T
        if self.bias is not None:
            result = result + self.bias
        return result

    def extra_repr(self):
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"



"""
Tests comparing energizer operators with PyTorch operators.
Tests mul, rmul, add, sub, and neg operations.
"""

import unittest
import numpy as np
import torch
import sys
import os

# Add parent directory to path to import energizer
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import energizer


class TestOperators(unittest.TestCase):
    """Test operators: mul, rmul, add, sub, and neg"""

    def setUp(self):
        """Set up test fixtures"""
        self.rtol = 1e-5  # Relative tolerance for floating point comparison
        self.atol = 1e-8  # Absolute tolerance for floating point comparison

    def assert_tensors_close(self, my_tensor, torch_tensor, msg=""):
        """Helper to compare energizer tensor with PyTorch tensor"""
        # Convert PyTorch tensor to numpy if needed
        if isinstance(torch_tensor, torch.Tensor):
            torch_data = torch_tensor.detach().numpy()
        else:
            torch_data = np.array(torch_tensor)

        # Compare data
        np.testing.assert_allclose(
            my_tensor.data,
            torch_data,
            rtol=self.rtol,
            atol=self.atol,
            err_msg=f"Tensor data mismatch. {msg}",
        )

    # ========== MUL TESTS ==========

    def test_mul_tensor_tensor_1d(self):
        """Test tensor * tensor for 1D tensors"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b = energizer.tensor([4.0, 5.0, 6.0])
        result = a * b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b_torch = torch.tensor([4.0, 5.0, 6.0])
        result_torch = a_torch * b_torch
        result_torch.sum().backward()
        self.assert_tensors_close(a.grad, a_torch.grad, "1D tensor * tensor")

    def test_rmul_tensor_scalar(self):
        """Test scalar * tensor"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = 2 * a
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = 2 * a_torch
        result_torch.sum().backward()
        self.assert_tensors_close(a.grad, a_torch.grad, "scalar * tensor")

    # ========== ADD TESTS ==========

    def test_add_tensor_tensor_1d(self):
        """Test tensor + tensor for 1D tensors"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b = energizer.tensor([4.0, 5.0, 6.0], requires_grad=True)
        result = a + b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b_torch = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
        result_torch = a_torch + b_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "1D tensor + tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "1D tensor + tensor grad a")
        self.assert_tensors_close(b.grad, b_torch.grad, "1D tensor + tensor grad b")

    def test_add_tensor_scalar(self):
        """Test tensor + scalar"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = a + 2.0
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = a_torch + 2.0
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "tensor + scalar forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "tensor + scalar grad")

    def test_add_tensor_tensor_2d(self):
        """Test tensor + tensor for 2D tensors"""
        # energizer
        a = energizer.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b = energizer.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result = a + b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b_torch = torch.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result_torch = a_torch + b_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "2D tensor + tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "2D tensor + tensor grad a")
        self.assert_tensors_close(b.grad, b_torch.grad, "2D tensor + tensor grad b")

    # ========== SUB TESTS ==========

    def test_sub_tensor_tensor_1d(self):
        """Test tensor - tensor for 1D tensors"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b = energizer.tensor([4.0, 5.0, 6.0], requires_grad=True)
        result = a - b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b_torch = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
        result_torch = a_torch - b_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "1D tensor - tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "1D tensor - tensor grad a")
        self.assert_tensors_close(b.grad, b_torch.grad, "1D tensor - tensor grad b")

    def test_sub_tensor_scalar(self):
        """Test tensor - scalar"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = a - 2.0
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = a_torch - 2.0
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "tensor - scalar forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "tensor - scalar grad")

    def test_sub_tensor_tensor_2d(self):
        """Test tensor - tensor for 2D tensors"""
        # energizer
        a = energizer.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b = energizer.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result = a - b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b_torch = torch.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result_torch = a_torch - b_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "2D tensor - tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "2D tensor - tensor grad a")
        self.assert_tensors_close(b.grad, b_torch.grad, "2D tensor - tensor grad b")

    # ========== NEG TESTS ==========

    def test_neg_tensor_1d(self):
        """Test -tensor for 1D tensors"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = -a
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = -a_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "1D -tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "1D -tensor grad")

    def test_neg_tensor_2d(self):
        """Test -tensor for 2D tensors"""
        # energizer
        a = energizer.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        result = -a
        result.backward()

        # PyTorch
        a_torch = torch.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        result_torch = -a_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "2D -tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "2D -tensor grad")

    def test_neg_tensor_negative_values(self):
        """Test -tensor with negative values"""
        # energizer
        a = energizer.tensor([-1.0, -2.0, 3.0], requires_grad=True)
        result = -a
        result.backward()

        # PyTorch
        a_torch = torch.tensor([-1.0, -2.0, 3.0], requires_grad=True)
        result_torch = -a_torch
        result_torch.sum().backward()

        self.assert_tensors_close(
            result, result_torch, "-tensor with negative values forward"
        )
        self.assert_tensors_close(
            a.grad, a_torch.grad, "-tensor with negative values grad"
        )

    # ========== MUL ADDITIONAL TESTS ==========

    def test_mul_tensor_scalar(self):
        """Test tensor * scalar"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = a * 2.0
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = a_torch * 2.0
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "tensor * scalar forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "tensor * scalar grad")

    def test_mul_tensor_tensor_2d(self):
        """Test tensor * tensor for 2D tensors"""
        # energizer
        a = energizer.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b = energizer.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result = a * b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([[1.0, 2.0], [3.0, 4.0]], requires_grad=True)
        b_torch = torch.tensor([[5.0, 6.0], [7.0, 8.0]], requires_grad=True)
        result_torch = a_torch * b_torch
        result_torch.sum().backward()

        self.assert_tensors_close(result, result_torch, "2D tensor * tensor forward")
        self.assert_tensors_close(a.grad, a_torch.grad, "2D tensor * tensor grad a")
        self.assert_tensors_close(b.grad, b_torch.grad, "2D tensor * tensor grad b")

    # ========== TRUEDIV TESTS ==========
    def test_truediv_tensor_tensor_1d(self):
        """Test tensor / tensor for 1D tensors"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b = energizer.tensor([4.0, 5.0, 6.0])
        result = a / b
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        b_torch = torch.tensor([4.0, 5.0, 6.0])
        result_torch = a_torch / b_torch
        result_torch.sum().backward()
        self.assert_tensors_close(a.grad, a_torch.grad, "1D tensor / tensor grad a")

    def test_truediv_tensor_scalar(self):
        """Test tensor / scalar"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = a / 2.0
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = a_torch / 2.0
        result_torch.sum().backward()
        self.assert_tensors_close(a.grad, a_torch.grad, "tensor / scalar grad")

    # ========== RTRUEDIV TESTS ==========
    def test_rtruediv_scalar_tensor(self):
        """Test scalar / tensor"""
        # energizer
        a = energizer.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result = 2.0 / a
        result.backward()

        # PyTorch
        a_torch = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        result_torch = 2.0 / a_torch
        result_torch.sum().backward()
        self.assert_tensors_close(a.grad, a_torch.grad, "scalar / tensor grad")


if __name__ == "__main__":
    unittest.main()



"""
backpack.runner
───────────────
CoreML inference engine for Energizer models.

Workflow:
    1. Train your model normally with Energizer (MLX/NumPy backend)
    2. Compile it once:  backpack.compile(model, example_input)  → .mlpackage
    3. Run it fast:      runner = backpack.runner.load("model.mlpackage")
                         output = runner(input_tensor)

The runner handles:
    - float32 ↔ float16 casting (ANE requires fp16)
    - numpy ↔ Energizer Tensor conversion
    - warm-up pass on load (avoids first-call JIT latency)
    - basic profiling (latency, throughput)
    - device visibility (which compute units were used)
"""

from __future__ import annotations

import time
import numpy as np
from dataclasses import dataclass, field
from typing import Optional, Union

# ── Types ─────────────────────────────────────────────────────────────────────


@dataclass
class RunnerConfig:
    """Configuration for the CoreML runner."""

    # Which compute units CoreML is allowed to use.
    # "ALL"        → CoreML decides (ANE + GPU + CPU), best for ANE dispatch
    # "CPU_AND_GPU"→ skips ANE (useful for debugging)
    # "CPU_ONLY"   → forces CPU only
    compute_units: str = "ALL"

    # Cast inputs to float16 before inference (required for ANE dispatch).
    # Set to False only if your model was compiled with float32 precision.
    fp16_inputs: bool = True

    # Number of warm-up passes on load (burns JIT compilation latency).
    warmup_runs: int = 2

    # If True, keep a rolling window of per-call latencies.
    track_latency: bool = True
    latency_window: int = 100


@dataclass
class RunStats:
    """Accumulated statistics for a runner session."""

    calls: int = 0
    total_ms: float = 0.0
    latencies_ms: list = field(default_factory=list)

    @property
    def avg_ms(self) -> float:
        return self.total_ms / self.calls if self.calls else 0.0

    @property
    def last_ms(self) -> float:
        return self.latencies_ms[-1] if self.latencies_ms else 0.0

    @property
    def p99_ms(self) -> float:
        if not self.latencies_ms:
            return 0.0
        s = sorted(self.latencies_ms)
        return s[int(len(s) * 0.99)]

    def record(self, ms: float, window: int):
        self.calls += 1
        self.total_ms += ms
        self.latencies_ms.append(ms)
        if len(self.latencies_ms) > window:
            self.latencies_ms.pop(0)

    def __repr__(self) -> str:
        return (
            f"RunStats(calls={self.calls}, "
            f"avg={self.avg_ms:.2f}ms, "
            f"last={self.last_ms:.2f}ms, "
            f"p99={self.p99_ms:.2f}ms)"
        )


# ── Runner ────────────────────────────────────────────────────────────────────


class Runner:
    """
    Wraps a compiled CoreML .mlpackage for fast inference.

    Usage:
        runner = Runner.load("my_model.mlpackage")
        output = runner(input_tensor)           # Energizer Tensor in/out
        output = runner.predict_numpy(arr)      # numpy in/out (lower overhead)
    """

    def __init__(
        self,
        model_path: str,
        config: Optional[RunnerConfig] = None,
    ):
        self._path = model_path
        self._config = config or RunnerConfig()
        self._model = None
        self._input_name: str = ""
        self._output_name: str = ""
        self._input_shape: tuple = ()
        self.stats = RunStats()

        self._load()

    # ── Loading ───────────────────────────────────────────────────────────────

    def _load(self):
        try:
            import coremltools as ct
        except ImportError:
            raise ImportError(
                "coremltools is required to use backpack.runner.\n"
                "Install it with:  pip install coremltools\n"
                "Note: requires Python ≤ 3.12."
            )

        # Map config string to CoreML compute units enum
        cu_map = {
            "ALL": ct.ComputeUnit.ALL,
            "CPU_AND_GPU": ct.ComputeUnit.CPU_AND_GPU,
            "CPU_ONLY": ct.ComputeUnit.CPU_ONLY,
        }
        cu = cu_map.get(self._config.compute_units.upper(), ct.ComputeUnit.ALL)

        self._model = ct.models.MLModel(self._path, compute_units=cu)

        spec = self._model.get_spec()
        self._input_name = spec.description.input[0].name
        self._output_name = spec.description.output[0].name

        # Infer input shape from spec
        feat = spec.description.input[0]
        if feat.type.HasField("multiArrayType"):
            self._input_shape = tuple(feat.type.multiArrayType.shape)

        # Warm-up: run dummy passes to burn CoreML's JIT compilation
        if self._config.warmup_runs > 0 and self._input_shape:
            dummy = np.zeros(
                self._input_shape,
                dtype=np.float16 if self._config.fp16_inputs else np.float32,
            )
            for _ in range(self._config.warmup_runs):
                self._model.predict({self._input_name: dummy})

    # ── Core predict ──────────────────────────────────────────────────────────

    def predict_numpy(self, x: np.ndarray) -> np.ndarray:
        """
        Run inference on a numpy array.
        Input is cast to fp16 if config.fp16_inputs is True.
        Output is always returned as float32 numpy array.
        """
        if self._config.fp16_inputs:
            x = x.astype(np.float16)
        else:
            x = x.astype(np.float32)

        t0 = time.perf_counter()
        result = self._model.predict({self._input_name: x})
        ms = (time.perf_counter() - t0) * 1000.0

        if self._config.track_latency:
            self.stats.record(ms, self._config.latency_window)

        out = result[self._output_name]
        return np.asarray(out, dtype=np.float32)

    def __call__(self, x) -> "Tensor":
        """
        Run inference on an Energizer Tensor.
        Returns an Energizer Tensor on CPU.
        """
        from energizer.tensor import Tensor

        if isinstance(x, Tensor):
            arr = x.numpy()
        elif isinstance(x, np.ndarray):
            arr = x
        else:
            arr = np.asarray(x, dtype=np.float32)

        out_arr = self.predict_numpy(arr)

        return Tensor(out_arr, device="cpu", requires_grad=False)

    # ── Batch inference ───────────────────────────────────────────────────────

    def predict_batch(self, inputs: list) -> list:
        """
        Run inference on a list of inputs (Tensors or numpy arrays).
        Returns a list of Energizer Tensors.
        Useful when CoreML model expects batch_size=1 (ANE constraint).
        """
        return [self(x) for x in inputs]

    # ── Utilities ─────────────────────────────────────────────────────────────

    @property
    def input_shape(self) -> tuple:
        return self._input_shape

    @property
    def input_name(self) -> str:
        return self._input_name

    @property
    def output_name(self) -> str:
        return self._output_name

    def benchmark(self, x, runs: int = 50) -> dict:
        """
        Run `runs` inference passes and return latency statistics.

        Returns:
            dict with keys: runs, avg_ms, min_ms, max_ms, p50_ms, p99_ms
        """
        if isinstance(x, type) and hasattr(x, "numpy"):
            arr = x.numpy()
        elif isinstance(x, np.ndarray):
            arr = x
        else:
            arr = np.asarray(x, dtype=np.float32)

        latencies = []
        for _ in range(runs):
            t0 = time.perf_counter()
            self.predict_numpy(arr)
            latencies.append((time.perf_counter() - t0) * 1000.0)

        latencies.sort()
        n = len(latencies)
        return {
            "runs": n,
            "avg_ms": sum(latencies) / n,
            "min_ms": latencies[0],
            "max_ms": latencies[-1],
            "p50_ms": latencies[n // 2],
            "p99_ms": latencies[int(n * 0.99)],
        }

    def describe(self) -> str:
        """Human-readable summary of the loaded model."""
        lines = [
            f"Runner  →  {self._path}",
            f"  input   : {self._input_name}  shape={self._input_shape}",
            f"  output  : {self._output_name}",
            f"  compute : {self._config.compute_units}",
            f"  fp16    : {self._config.fp16_inputs}",
            f"  warmup  : {self._config.warmup_runs} runs",
            f"  stats   : {self.stats}",
        ]
        return "\n".join(lines)

    def __repr__(self) -> str:
        return f"Runner(path='{self._path}', compute='{self._config.compute_units}')"

    # ── Class-level factory ───────────────────────────────────────────────────

    @classmethod
    def load(
        cls,
        path: str,
        compute_units: str = "ALL",
        fp16_inputs: bool = True,
        warmup_runs: int = 2,
    ) -> "Runner":
        """
        Convenience factory.

        Example:
            runner = Runner.load("model.mlpackage")
            output = runner(my_tensor)
        """
        config = RunnerConfig(
            compute_units=compute_units,
            fp16_inputs=fp16_inputs,
            warmup_runs=warmup_runs,
        )
        return cls(path, config)


# ── Module-level convenience ──────────────────────────────────────────────────


def load(
    path: str,
    compute_units: str = "ALL",
    fp16_inputs: bool = True,
    warmup_runs: int = 2,
) -> Runner:
    """
    Load a compiled .mlpackage and return a ready-to-use Runner.

    Args:
        path:          Path to the .mlpackage directory.
        compute_units: "ALL" (default) | "CPU_AND_GPU" | "CPU_ONLY"
        fp16_inputs:   Cast inputs to float16 (required for ANE, default True).
        warmup_runs:   Number of warm-up inference passes on load (default 2).

    Returns:
        Runner instance.

    Example:
        import backpack.runner as runner

        model  = runner.load("my_model.mlpackage")
        output = model(input_tensor)
        print(model.stats)
    """
    return Runner.load(
        path,
        compute_units=compute_units,
        fp16_inputs=fp16_inputs,
        warmup_runs=warmup_runs,
    )


def benchmark(
    path: str,
    example_input: np.ndarray,
    runs: int = 50,
    compute_units: str = "ALL",
) -> dict:
    """
    One-shot benchmark of a compiled model.

    Example:
        import numpy as np
        import backpack.runner as runner

        stats = runner.benchmark("model.mlpackage", np.zeros((1, 784)))
        print(f"avg latency: {stats['avg_ms']:.2f} ms")
    """
    r = load(path, compute_units=compute_units)
    return r.benchmark(example_input, runs=runs)



from energizer.neural_network import Module, Parameter
import energizer
import numpy as np
import mlx.core as mx
from energizer.tensor import Tensor


class ConvTranspose2d(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple,
        stride: tuple = 1,
        padding: tuple = 0,
        output_padding: tuple = 0,
        groups: int = 1,
        bias: bool = True,
        dilation: tuple = 1,
        device: str = "cpu",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = (
            kernel_size
            if isinstance(kernel_size, tuple)
            else (kernel_size, kernel_size)
        )
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.output_padding = (
            output_padding
            if isinstance(output_padding, tuple)
            else (output_padding, output_padding)
        )
        self.groups = groups
        self.bias = bias
        self.dilation = (
            dilation if isinstance(dilation, tuple) else (dilation, dilation)
        )
        self.device = device

        weight_shape = (
            in_channels,
            out_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        )

        if device == "gpu":
            weight_data = mx.random.normal(weight_shape) * 0.01
            bias_data = mx.zeros(out_channels) if bias else None
        else:
            weight_data = np.random.randn(*weight_shape) * 0.01
            bias_data = np.zeros(out_channels) if bias else None

        self.weight = Parameter(weight_data, requires_grad=True, device=device)

        if bias:
            self.bias = Parameter(bias_data, requires_grad=True, device=device)
        else:
            self.bias = None

    def forward(self, x: Tensor) -> Tensor:
        if x.device != self.device:
            x = x.to(self.device)

        batch_size, in_channels, height, width = x.shape

        kh, kw = self.kernel_size
        sh, sw = self.stride
        ph, pw = self.padding
        oph, opw = self.output_padding
        dh, dw = self.dilation

        if (dh, dw) != (1, 1):
            raise NotImplementedError(
                "ConvTranspose2d dilation != 1 is not supported yet"
            )

        if self.device == "gpu":
            if not isinstance(x.data, mx.array):
                x_data = mx.array(x.data)
            else:
                x_data = x.data

            if not isinstance(self.weight.data, mx.array):
                weight_data = mx.array(self.weight.data)
            else:
                weight_data = self.weight.data

            x_channels_last = mx.transpose(x_data, (0, 2, 3, 1))

            weight_mlx = mx.transpose(weight_data, (1, 2, 3, 0))

            kwargs = {
                "stride": (sh, sw),
                "padding": (ph, pw),
                "dilation": (dh, dw),
                "groups": self.groups,
            }
            if (oph, opw) != (0, 0):
                kwargs["output_padding"] = (oph, opw)

            output = mx.conv_transpose2d(x_channels_last, weight_mlx, **kwargs)
            output = mx.transpose(output, (0, 3, 1, 2))
        else:
            x_data = x.data if isinstance(x.data, np.ndarray) else np.array(x.data)
            weight_data = (
                self.weight.data
                if isinstance(self.weight.data, np.ndarray)
                else np.array(self.weight.data)
            )

            out_height = (height - 1) * sh - 2 * ph + (kh - 1) + oph + 1
            out_width = (width - 1) * sw - 2 * pw + (kw - 1) + opw + 1

            up_h = (height - 1) * sh + 1
            up_w = (width - 1) * sw + 1
            upsampled = np.zeros(
                (batch_size, in_channels, up_h, up_w), dtype=x_data.dtype
            )
            upsampled[:, :, ::sh, ::sw] = x_data

            pad_h = kh - 1 - ph
            pad_w = kw - 1 - pw
            padded = np.pad(
                upsampled,
                ((0, 0), (0, 0), (pad_h, pad_h), (pad_w, pad_w)),
                mode="constant",
            )

            flipped_weight = np.flip(weight_data, axis=(2, 3))
            conv_h = padded.shape[2] - kh + 1
            conv_w = padded.shape[3] - kw + 1

            output = np.zeros(
                (batch_size, self.out_channels, conv_h, conv_w), dtype=x_data.dtype
            )

            for b in range(batch_size):
                for oc in range(self.out_channels):
                    for ic in range(in_channels):
                        kernel = flipped_weight[ic, oc]
                        for i in range(conv_h):
                            for j in range(conv_w):
                                patch = padded[b, ic, i : i + kh, j : j + kw]
                                output[b, oc, i, j] += np.sum(patch * kernel)

            if oph or opw:
                output = np.pad(
                    output, ((0, 0), (0, 0), (0, oph), (0, opw)), mode="constant"
                )

            output = output[:, :, :out_height, :out_width]

        if self.bias is not None:
            if self.device == "gpu":
                if isinstance(output, mx.array):
                    bias_data = (
                        self.bias.mlx()
                        if hasattr(self.bias, "mlx")
                        else mx.array(self.bias.data)
                    )
                    output += bias_data.reshape(1, -1, 1, 1)
                else:
                    output += np.array(self.bias.data).reshape(1, -1, 1, 1)
            else:
                bias_data = (
                    self.bias.data
                    if isinstance(self.bias.data, np.ndarray)
                    else np.array(self.bias.data)
                )
                output += bias_data.reshape(1, -1, 1, 1)

        return Tensor(output, requires_grad=x.requires_grad, device=self.device)

    def to(self, device: str):
        self.device = device
        self.weight = self.weight.to(device)
        if self.bias is not None:
            self.bias = self.bias.to(device)
        return self



from energizer.neural_network import Module
from energizer.tensor import Tensor
import energizer.autograd as autograd


class Reshape(Module):
    def __init__(self, shape: tuple):
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        if not isinstance(x, Tensor):
            raise TypeError(f"Expected Tensor, got {type(x)}")

        new_shape = list(self.shape)

        if -1 in new_shape:
            total_elements = np.prod(x.shape)
            known_elements = np.prod([abs(d) for d in new_shape if d != -1])
            inferred_dim = total_elements // known_elements
            new_shape[new_shape.index(-1)] = inferred_dim

        new_shape = tuple(new_shape)

        return autograd.Reshape.apply(x, new_shape)



from energizer.neural_network import Module
import energizer


class Sequential(Module):
    def __init__(self, *layers):
        super().__init__()
        self.layers = layers
        for idx, layer in enumerate(layers):
            self.add_module(str(idx), layer)

    def forward(self, x: energizer.Tensor) -> energizer.Tensor:
        for layer in self.layers:
            x = layer(x)
        return x

    def to(self, device: str):
        for layer in self.layers:
            layer.to(device)
        return self



from energizer.neural_network import Module, Parameter
from energizer.tensor import Tensor
import numpy as np

try:
    import mlx.core as mx
except ImportError:
    mx = None


class BatchNorm1d(Module):
    def __init__(self, num_features: int, eps: float = 1e-5, momentum: float = 0.1):
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum

        self.gamma = Tensor(np.zeros(num_features), requires_grad=True)
        self.beta = Tensor(np.zeros(num_features), requires_grad=True)

        self.running_mean = np.zeros(num_features)
        self.running_var = np.ones(num_features)

        self.training = True

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            if len(x.data.shape) == 2:
                mean = x.data.mean(axis=0)
                var = x.data.var(axis=0)
            else:
                mean = x.data.mean(axis=(0, 2))
                var = x.data.var(axis=(0, 2))

            self.running_mean = (
                1 - self.momentum
            ) * self.running_mean + self.momentum * mean
            self.running_var = (
                1 - self.momentum
            ) * self.running_var + self.momentum * var

            x_normalized = (x.data - mean) / np.sqrt(var + self.eps)

        else:
            x_normalized = (x.data - self.running_mean) / np.sqrt(
                self.running_var + self.eps
            )

        if len(x.data.shape) == 2:
            result = x_normalized * self.gamma + self.beta
        else:
            gamma_reshaped = self.gamma.data.reshape(1, -1, 1)
            beta_reshaped = self.beta.data.reshape(1, -1, 1)
            result = x_normalized * gamma_reshaped + beta_reshaped

        return Tensor(result, requires_grad=x.requires_grad)

    def eval(self):
        self.training = False

    def train(self):
        self.training = True

    def parameters(self):
        return [self.gamma, self.beta]


class BatchNorm2d(Module):
    def __init__(self, num_features: int, eps: float = 1e-5, momentum: float = 0.1):
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum

        self.gamma = Parameter(np.ones(num_features), requires_grad=True)
        self.beta = Parameter(np.zeros(num_features), requires_grad=True)

        self.running_mean = np.zeros(num_features)
        self.running_var = np.ones(num_features)

        self.training = True

    def forward(self, x: Tensor) -> Tensor:
        if len(x.data.shape) != 4:
            raise ValueError("BatchNorm2d only supports 4D input tensors")

        batch_size, channels, height, width = x.data.shape

        if self.training:
            mean = x.data.mean(axis=(0, 2, 3))
            var = x.data.var(axis=(0, 2, 3))

            self.running_mean = (
                1 - self.momentum
            ) * self.running_mean + self.momentum * mean
            self.running_var = (
                1 - self.momentum
            ) * self.running_var + self.momentum * var

            # reshape for broadcasting
            mean_reshaped = mean.reshape(1, channels, 1, 1)
            var_reshaped = var.reshape(1, channels, 1, 1)

        else:
            mean_reshaped = self.running_mean.reshape(1, channels, 1, 1)
            var_reshaped = self.running_var.reshape(1, channels, 1, 1)

        x_normalized = (x.data - mean_reshaped) / (var_reshaped + self.eps) ** 0.5

        gamma_reshaped = self.gamma.data.reshape(1, channels, 1, 1)
        beta_reshaped = self.beta.data.reshape(1, channels, 1, 1)

        result = x_normalized * gamma_reshaped + beta_reshaped

        return Tensor(result, requires_grad=x.requires_grad, device=x.device)

    def to(self, device: str):
        super().to(device)
        if device == "gpu" and mx is not None:
            if isinstance(self.running_mean, np.ndarray):
                self.running_mean = mx.array(self.running_mean)
            if isinstance(self.running_var, np.ndarray):
                self.running_var = mx.array(self.running_var)
        else:
            if not isinstance(self.running_mean, np.ndarray):
                self.running_mean = np.array(self.running_mean)
            if not isinstance(self.running_var, np.ndarray):
                self.running_var = np.array(self.running_var)
        return self

    def eval(self):
        self.training = False

    def train(self):
        self.training = True
